import os
from typing import Optional

import torch

from chebai.preprocessing.datasets.base import XYBaseDataModule
from chebai.preprocessing.datasets.chebi import _ChEBIDataExtractor


class BCEWeighted(torch.nn.BCEWithLogitsLoss):
    """
    BCEWithLogitsLoss with weights automatically computed according to the beta parameter.

    This class computes weights based on the formula from the paper by Cui et al. (2019):
    https://openaccess.thecvf.com/content_CVPR_2019/papers/Cui_Class-Balanced_Loss_Based_on_Effective_Number_of_Samples_CVPR_2019_paper.pdf

    Args:
        beta (float, optional): The beta parameter for weight calculation. Default is None.
        data_extractor (XYBaseDataModule, optional): The data extractor for loading the dataset. Default is None.
    """

    def __init__(
        self,
        beta: float = 0.99,
        data_extractor: Optional[XYBaseDataModule] = None,
        **kwargs,
    ):
        from chebai.preprocessing.datasets.pubchem import LabeledUnlabeledMixed

        self.beta = beta
        if isinstance(data_extractor, LabeledUnlabeledMixed):
            data_extractor = data_extractor.labeled
        self.data_extractor = data_extractor

        assert (
            isinstance(beta, float) and beta > 0.0
        ), f"Beta parameter must be a float with value greater than 0.0, for loss class {self.__class__.__name__}."

        assert (
            self.data_extractor is not None
        ), f"Data extractor must be provided if this loss class ({self.__class__.__name__}) is used."

        assert all(
            os.path.exists(os.path.join(self.data_extractor.processed_dir, file_name))
            for file_name in self.data_extractor.processed_file_names
        ), "Dataset files not found. Make sure the dataset is processed before using this loss."

        assert (
            isinstance(self.data_extractor, _ChEBIDataExtractor)
            or self.data_extractor is None
        )
        super().__init__(**kwargs)
        self.pos_weight: Optional[torch.Tensor] = None

    def set_pos_weight(self, input: torch.Tensor) -> None:
        """
        Sets the positive weights for the loss function based on the input tensor.

        Args:
            input (torch.Tensor): The input tensor for which to set the positive weights.
        """
        if self.pos_weight is None:
            print(
                f"Computing loss-weights based on v{self.data_extractor.chebi_version} dataset (beta={self.beta})"
            )
            complete_labels = torch.concat(
                [
                    torch.stack(
                        [
                            torch.Tensor(row["labels"])
                            for row in self.data_extractor.load_processed_data(
                                filename=file_name
                            )
                        ]
                    )
                    for file_name in self.data_extractor.processed_file_names
                ]
            )
            value_counts = complete_labels.sum(dim=0)
            weights = [
                (1 - self.beta) / (1 - pow(self.beta, value)) for value in value_counts
            ]
            mean = sum(weights) / len(weights)
            self.pos_weight = torch.tensor(
                [w / mean for w in weights], device=input.device
            )

    def forward(
        self, input: torch.Tensor, target: torch.Tensor, **kwargs
    ) -> torch.Tensor:
        """
        Forward pass for the loss calculation.

        Args:
            input (torch.Tensor): The input tensor (predictions).
            target (torch.Tensor): The target tensor (labels).

        Returns:
            torch.Tensor: The computed loss.
        """
        self.set_pos_weight(input)
        return super().forward(input, target)


class UnWeightedBCEWithLogitsLoss(torch.nn.BCEWithLogitsLoss):
    def forward(self, input, target, **kwargs):
        # As the custom passed kwargs are not used in BCEWithLogitsLoss, we can ignore them
        return super().forward(input, target)



import os
from datetime import datetime
from typing import List, Literal, Optional, Union

from lightning.fabric.utilities.types import _PATH
from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import WandbLogger


class CustomLogger(WandbLogger):
    """
    A custom logger that extends WandbLogger to add support for custom naming of runs and cross-validation.

    Args:
        save_dir (_PATH): Directory where logs are saved.
        name (str): Name of the logging run.
        version (Optional[Union[int, str]]): Version of the logging run.
        prefix (str): Prefix for logging.
        fold (Optional[int]): Cross-validation fold number.
        project (Optional[str]): Wandb project name.
        entity (Optional[str]): Wandb entity name.
        offline (bool): Whether to log offline.
        log_model (Union[Literal["all"], bool]): Whether to log the model.
        verbose_hyperparameters (bool): Whether to log hyperparameters verbosely.
        tags (Optional[List[str]]): List of tags for the run.
        **kwargs: Additional keyword arguments for WandbLogger.
    """

    def __init__(
        self,
        save_dir: _PATH,
        name: str = "logs",
        version: Optional[Union[int, str]] = None,
        prefix: str = "",
        fold: Optional[int] = None,
        project: Optional[str] = None,
        entity: Optional[str] = None,
        offline: bool = False,
        log_model: Union[Literal["all"], bool] = False,
        verbose_hyperparameters: bool = False,
        tags: Optional[List[str]] = None,
        **kwargs,
    ):
        if version is None:
            version = f"{datetime.now():%y%m%d-%H%M}"
        self._version = version
        self._name = name
        self._fold = fold
        self.verbose_hyperparameters = verbose_hyperparameters
        super().__init__(
            name=self.name,
            save_dir=save_dir,
            version=None,
            prefix=prefix,
            log_model=log_model,
            entity=entity,
            project=project,
            offline=offline,
            **kwargs,
        )
        if tags:
            self.experiment.tags += tuple(tags)

    @property
    def name(self) -> Optional[str]:
        """
        Returns the name of the logging run, including the version and fold number if applicable.
        """
        name = f"{self._name}_{self.version}"
        if self._fold is not None:
            name += f"_fold{self._fold}"
        return name

    @property
    def version(self) -> Optional[str]:
        """
        Returns the version of the logging run.
        """
        return self._version

    @property
    def root_dir(self) -> Optional[str]:
        """
        Returns the root directory for saving logs.
        """
        return os.path.join(self.save_dir, self.name)

    @property
    def log_dir(self) -> str:
        """
        Returns the directory for saving logs, including the version and fold number if applicable.
        """
        version = (
            self.version if isinstance(self.version, str) else f"version_{self.version}"
        )
        if self._fold is None:
            return os.path.join(self.root_dir, version)
        return os.path.join(self.root_dir, version, f"fold_{self._fold}")

    def set_fold(self, fold: int) -> None:
        """
        Sets the fold number and restarts the Wandb experiment with the new fold number.

        Args:
            fold (int): Cross-validation fold number.
        """
        import wandb

        if fold != self._fold:
            self._fold = fold
            # Start new experiment
            wandb.finish()
            self._wandb_init["name"] = self.name
            self._experiment = None
            _ = self.experiment

    @property
    def fold(self) -> Optional[int]:
        """
        Returns the current fold number.
        """
        return self._fold

    def after_save_checkpoint(self, checkpoint_callback: ModelCheckpoint) -> None:
        """
        Override method to prevent saving checkpoints as Wandb artifacts.
        """
        pass



import torch
import torch.nn as nn
import torch.nn.functional as F


# from https://github.com/itakurah/Focal-loss-PyTorch


class FocalLoss(nn.Module):
    def __init__(
        self,
        gamma=2,
        alpha=None,
        reduction="mean",
        task_type="binary",
        num_classes=None,
    ):
        """
        Unified Focal Loss class for binary, multi-class, and multi-label classification tasks.
        :param gamma: Focusing parameter, controls the strength of the modulating factor (1 - p_t)^gamma
        :param alpha: Balancing factor, can be a scalar or a tensor for class-wise weights. If None, no class balancing is used.
        :param reduction: Specifies the reduction method: 'none' | 'mean' | 'sum'
        :param task_type: Specifies the type of task: 'binary', 'multi-class', or 'multi-label'
        :param num_classes: Number of classes (only required for multi-class classification)
        """
        super(FocalLoss, self).__init__()
        self.gamma = gamma
        self.alpha = alpha
        self.reduction = reduction
        self.task_type = task_type
        self.num_classes = num_classes

        # Handle alpha for class balancing in multi-class tasks
        if (
            task_type == "multi-class"
            and alpha is not None
            and isinstance(alpha, (list, torch.Tensor))
        ):
            assert (
                num_classes is not None
            ), "num_classes must be specified for multi-class classification"
            if isinstance(alpha, list):
                self.alpha = torch.Tensor(alpha)
            else:
                self.alpha = alpha

    def forward(self, inputs, targets):
        """
        Forward pass to compute the Focal Loss based on the specified task type.
        :param inputs: Predictions (logits) from the model.
                       Shape:
                         - binary/multi-label: (batch_size, num_classes)
                         - multi-class: (batch_size, num_classes)
        :param targets: Ground truth labels.
                        Shape:
                         - binary: (batch_size,)
                         - multi-label: (batch_size, num_classes)
                         - multi-class: (batch_size,)
        """
        if self.task_type == "binary":
            return self.binary_focal_loss(inputs, targets)
        elif self.task_type == "multi-class":
            return self.multi_class_focal_loss(inputs, targets)
        elif self.task_type == "multi-label":
            return self.multi_label_focal_loss(inputs, targets)
        else:
            raise ValueError(
                f"Unsupported task_type '{self.task_type}'. Use 'binary', 'multi-class', or 'multi-label'."
            )

    def binary_focal_loss(self, inputs, targets):
        """Focal loss for binary classification."""
        probs = torch.sigmoid(inputs)
        targets = targets.float()

        # Compute binary cross entropy
        bce_loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction="none")

        # Compute focal weight
        p_t = probs * targets + (1 - probs) * (1 - targets)
        focal_weight = (1 - p_t) ** self.gamma

        # Apply alpha if provided
        if self.alpha is not None:
            alpha_t = self.alpha * targets + (1 - self.alpha) * (1 - targets)
            bce_loss = alpha_t * bce_loss

        # Apply focal loss weighting
        loss = focal_weight * bce_loss

        if self.reduction == "mean":
            return loss.mean()
        elif self.reduction == "sum":
            return loss.sum()
        return loss

    def multi_class_focal_loss(self, inputs, targets):
        """Focal loss for multi-class classification."""
        if self.alpha is not None:
            alpha = self.alpha.to(inputs.device)

        # Convert logits to probabilities with softmax
        probs = F.softmax(inputs, dim=1)

        # One-hot encode the targets
        targets_one_hot = F.one_hot(targets, num_classes=self.num_classes).float()

        # Compute cross-entropy for each class
        ce_loss = -targets_one_hot * torch.log(probs)

        # Compute focal weight
        p_t = torch.sum(probs * targets_one_hot, dim=1)  # p_t for each sample
        focal_weight = (1 - p_t) ** self.gamma

        # Apply alpha if provided (per-class weighting)
        if self.alpha is not None:
            alpha_t = alpha.gather(0, targets)
            ce_loss = alpha_t.unsqueeze(1) * ce_loss

        # Apply focal loss weight
        loss = focal_weight.unsqueeze(1) * ce_loss

        if self.reduction == "mean":
            return loss.mean()
        elif self.reduction == "sum":
            return loss.sum()
        return loss

    def multi_label_focal_loss(self, inputs, targets):
        """Focal loss for multi-label classification."""
        probs = torch.sigmoid(inputs)

        # Compute binary cross entropy
        bce_loss = F.binary_cross_entropy_with_logits(inputs, targets, reduction="none")

        # Compute focal weight
        p_t = probs * targets + (1 - probs) * (1 - targets)
        focal_weight = (1 - p_t) ** self.gamma

        # Apply alpha if provided
        if self.alpha is not None:
            alpha_t = self.alpha * targets + (1 - self.alpha) * (1 - targets)
            bce_loss = alpha_t * bce_loss

        # Apply focal loss weight
        loss = focal_weight * bce_loss

        if self.reduction == "mean":
            return loss.mean()
        elif self.reduction == "sum":
            return loss.sum()
        return loss



import json
from pathlib import Path
from typing import Literal

import torchmetrics
from jsonargparse import CLI
from torchmetrics.classification import MultilabelConfusionMatrix, MultilabelF1Score

from chebai.preprocessing.datasets.base import XYBaseDataModule
from chebai.result.utils import (
    load_data_instance,
    load_model_for_inference,
    parse_config_file,
)


class ClassesPropertiesGenerator:
    """
    Computes PPV (Positive Predictive Value) and NPV (Negative Predictive Value) and counts the number of
    true positives (TP), false positives (FP), true negatives (TN), and false negatives (FN)
    for each class in a multi-label classification problem using a PyTorch Lightning model.
    """

    @staticmethod
    def load_class_labels(path: Path) -> list[str]:
        """
        Load a list of class names from a .json or .txt file.

        Args:
            path: Path to the class labels file (txt or json).

        Returns:
            A list of class names, one per line.
        """
        path = Path(path)
        with path.open() as f:
            return [line.strip() for line in f if line.strip()]

    @staticmethod
    def compute_classwise_scores(
        metrics_obj_dict: dict[str, torchmetrics.Metric],
        class_names: list[str],
    ) -> dict[str, dict[str, float]]:
        """
        Compute per-class evaluation metrics for a multi-label classification task.

        This method uses torchmetrics objects (MultilabelConfusionMatrix, F1 scores, etc.)
        to compute the following metrics for each class:
            - PPV (Positive Predictive Value or Precision)
            - NPV (Negative Predictive Value)
            - True Positives (TP)
            - False Positives (FP)
            - True Negatives (TN)
            - False Negatives (FN)
            - F1 score

        Args:
            metrics_obj_dict: Dictionary containing pre-updated torchmetrics.Metric objects:
                {
                    "cm": MultilabelConfusionMatrix,
                    "micro-f1": MultilabelF1Score (average=None)
                }
            class_names: List of class names in the same order as class indices.

        Returns:
            Dictionary mapping each class name to a sub-dictionary of computed metrics:
            {
                "class_name_1": {
                    "PPV": float,
                    "NPV": float,
                    "TN": int,
                    "FP": int,
                    "FN": int,
                    "TP": int,
                    "f1": float,
                },
                ...
            }
        """
        cm_tensor = metrics_obj_dict["cm"].compute()  # Shape: (num_classes, 2, 2)
        f1_tensor = metrics_obj_dict["f1"].compute()  # shape: (num_classes,)

        assert len(class_names) == cm_tensor.shape[0] == f1_tensor.shape[0], (
            f"Mismatch between number of class names ({len(class_names)}) and metric tensor sizes: "
            f"confusion matrix has {cm_tensor.shape[0]}, "
            f"F1 has {f1_tensor.shape[0]}, "
        )

        results: dict[str, dict[str, float]] = {}
        for idx, cls_name in enumerate(class_names):
            tn = cm_tensor[idx][0][0].item()
            fp = cm_tensor[idx][0][1].item()
            fn = cm_tensor[idx][1][0].item()
            tp = cm_tensor[idx][1][1].item()

            ppv = tp / (tp + fp) if (tp + fp) > 0 else 0.0  # Precision
            npv = tn / (tn + fn) if (tn + fn) > 0 else 0.0  # Negative predictive value

            # positive_raw = [p.item() for i, p in enumerate(raw_preds[:, idx]) if true_np[i, idx]]
            # negative_raw = [p.item() for i, p in enumerate(raw_preds[:, idx]) if not true_np[i, idx]]

            f1 = f1_tensor[idx]

            results[cls_name] = {
                "PPV": round(ppv, 4),
                "NPV": round(npv, 4),
                "TN": int(tn),
                "FP": int(fp),
                "FN": int(fn),
                "TP": int(tp),
                "f1": round(f1.item(), 4),
                # "positive_preds": positive_raw,
                # "negative_preds": negative_raw,
            }
        return results

    def generate_props(
        self,
        data_partition: Literal["train", "val", "test"],
        model_ckpt_path: str,
        model_config_file_path: str,
        data_config_file_path: str,
        output_path: str | None = None,
        apply_id_filter: str | None = None,
    ) -> None:
        """
        Run inference on validation set, compute TPV/NPV  per class, and save to JSON.

        Args:
            data_partition: Partition of the dataset to use to generate class properties.
            model_ckpt_path: Path to the PyTorch Lightning checkpoint file.
            model_config_file_path: Path to yaml config file of the model.
            data_config_file_path: Path to yaml config file of the data.
            output_path: Optional path where to write the JSON metrics file.
                         Defaults to '<processed_dir_main>/classes.json'.
            apply_id_filter: Optional path to a (data.pt) file containing IDs to filter the dataset. This is useful for comparing datasets with different ids.
        """
        data_cls_path, data_cls_kwargs = parse_config_file(data_config_file_path)
        data_module: XYBaseDataModule = load_data_instance(
            data_cls_path, data_cls_kwargs
        )
        data_module.apply_id_filter = apply_id_filter

        splits_file_path = Path(data_module.processed_dir_main, "splits.csv")
        if data_module.splits_file_path is None:
            if not splits_file_path.exists():
                raise RuntimeError(
                    "Either the data module should be initialized with a `splits_file_path`, "
                    f"or the file `{splits_file_path}` must exists.\n"
                    "This is to prevent the data module from dynamically generating the splits."
                )

            print(
                f"`splits_file_path` is not provided as an initialization parameter to the data module\n"
                f"Using splits from the file {splits_file_path}"
            )
            data_module.splits_file_path = splits_file_path

        model_class_path, model_kwargs = parse_config_file(model_config_file_path)
        model = load_model_for_inference(
            model_ckpt_path, model_class_path, model_kwargs
        )

        if data_partition == "train":
            data_loader = data_module.train_dataloader()
        elif data_partition == "val":
            data_loader = data_module.val_dataloader()
        elif data_partition == "test":
            data_loader = data_module.test_dataloader()
        else:
            raise ValueError(f"Unknown data partition: {data_partition}")
        print(f"Running inference on {data_partition} data...")

        if data_module.apply_label_filter is not None:
            classes_file = data_module.apply_label_filter
        else:
            classes_file = Path(data_module.processed_dir_main) / "classes.txt"
        class_names = self.load_class_labels(classes_file)
        num_classes = len(class_names)
        metrics_obj_dict: dict[str, torchmetrics.Metric] = {
            "cm": MultilabelConfusionMatrix(num_labels=num_classes).to(
                device=model.device
            ),
            "f1": MultilabelF1Score(num_labels=num_classes, average=None).to(
                device=model.device
            ),
        }

        for batch_idx, batch in enumerate(data_loader):
            batch = batch.to(device=model.device)
            data = model._process_batch(batch, batch_idx=batch_idx)
            labels = data["labels"].to(device=model.device)
            data["features"][0].to(device=model.device)
            model_output = model(data, **data.get("model_kwargs", {}))
            preds, targets = model._get_prediction_and_labels(
                data, labels, model_output
            )
            for metric_obj in metrics_obj_dict.values():
                metric_obj.update(preds, targets)

        print("Computing metrics...")
        if output_path is None:
            output_file = (
                Path(data_module.processed_dir_main) / f"classes_{data_partition}.json"
            )
        else:
            output_file = Path(output_path)

        metrics = self.compute_classwise_scores(metrics_obj_dict, class_names)

        with output_file.open("w") as f:
            json.dump(metrics, f, indent=2)
        print(f"Saved metrics to {output_file}")


class Main:
    """
    CLI wrapper for ClassesPropertiesGenerator.
    """

    def generate(
        self,
        data_partition: Literal["train", "val", "test"],
        model_ckpt_path: str,
        model_config_file_path: str,
        data_config_file_path: str,
        output_path: str | None = None,
        apply_id_filter: str | None = None,
    ) -> None:
        """
        CLI command to generate JSON with metrics on validation set.

        Args:
            data_partition: Partition of dataset to use to generate class properties.
            model_ckpt_path: Path to the PyTorch Lightning checkpoint file.
            model_config_file_path: Path to yaml config file of the model.
            data_config_file_path: Path to yaml config file of the data.
            output_path: Optional path where to write the JSON metrics file.
                         Defaults to '<processed_dir_main>/classes.json'.
        """
        assert data_partition in [
            "train",
            "val",
            "test",
        ], f"Given data partition invalid: {data_partition}, Choose one of the value among `train`, `val`, `test` "
        generator = ClassesPropertiesGenerator()
        generator.generate_props(
            data_partition,
            model_ckpt_path,
            model_config_file_path,
            data_config_file_path,
            output_path,
            apply_id_filter=apply_id_filter,
        )


if __name__ == "__main__":
    # Usage:
    # generate_classes_properties.py generate \
    # --data_partition "val" \
    # --model_ckpt_path "model/ckpt/path" \
    # --model_config_file_path "model/config/file/path" \
    # --data_config_file_path "data/config/file/path" \
    # --output_path "output/file/path" # Optional
    CLI(Main, as_positional=False)



from collections import OrderedDict
from typing import Any, Dict, List, Optional, Tuple

import torch
from torch import Tensor, nn

from chebai.models import ChebaiBaseNet
from chebai.models.electra import filter_dict


class FFN(ChebaiBaseNet):
    # Reference: https://github.com/bio-ontology-research-group/deepgo2/blob/main/deepgo/models.py#L121-L139

    def __init__(
        self,
        hidden_layers: List[int] = [
            1024,
        ],
        use_adam_optimizer: bool = False,
        pretrained_checkpoint: Optional[str] = None,
        load_prefix: Optional[str] = "model.",
        **kwargs,
    ):
        super().__init__(**kwargs)

        self.use_adam_optimizer: bool = bool(use_adam_optimizer)
        print(f"Using Adam optimizer: {self.use_adam_optimizer}")

        layers = []
        current_layer_input_size = self.input_dim
        for hidden_dim in hidden_layers:
            layers.append(MLPBlock(current_layer_input_size, hidden_dim))
            layers.append(Residual(MLPBlock(hidden_dim, hidden_dim)))
            current_layer_input_size = hidden_dim

        layers.append(torch.nn.Linear(current_layer_input_size, self.out_dim))
        self.model = nn.Sequential(*layers)

        if pretrained_checkpoint is not None:
            ckpt_file = torch.load(
                pretrained_checkpoint, map_location=self.device, weights_only=False
            )
            if load_prefix is not None:
                state_dict = filter_dict(ckpt_file["state_dict"], load_prefix)
            else:
                state_dict = ckpt_file["state_dict"]

            model_sd = self.model.state_dict()
            filtered = OrderedDict()
            skipped = set()
            for k, v in state_dict.items():
                if model_sd[k].shape == v.shape:
                    filtered[k] = v  # only load params with matching shapes
                else:
                    skipped.add(k)
                    filtered[k] = model_sd[k]
                # else: silently skip mismatched keys like "2.weight", "2.bias"
                # which is the last linear layers which maps to output dimension

            self.model.load_state_dict(filtered)
            print(
                f"Loaded (shape-matched) weights from {pretrained_checkpoint}",
                f"Skipped the following weights: {skipped}",
            )

    def _get_prediction_and_labels(self, data, labels, model_output):
        d = model_output["logits"]
        loss_kwargs = data.get("loss_kwargs", dict())
        if "non_null_labels" in loss_kwargs:
            n = loss_kwargs["non_null_labels"]
            d = d[n]
        return torch.sigmoid(d), labels.int() if labels is not None else None

    def _process_for_loss(
        self,
        model_output: Dict[str, Tensor],
        labels: Tensor,
        loss_kwargs: Dict[str, Any],
    ) -> Tuple[Tensor, Tensor, Dict[str, Any]]:
        """
        Process the model output for calculating the loss.

        Args:
            model_output (Dict[str, Tensor]): The output of the model.
            labels (Tensor): The target labels.
            loss_kwargs (Dict[str, Any]): Additional loss arguments.

        Returns:
            tuple: A tuple containing the processed model output, labels, and loss arguments.
        """
        kwargs_copy = dict(loss_kwargs)
        if labels is not None:
            labels = labels.float()
        return model_output["logits"], labels, kwargs_copy

    def forward(self, data, **kwargs):
        x = data["features"]
        return {"logits": self.model(x)}

    def configure_optimizers(self, **kwargs) -> torch.optim.Optimizer:
        """
        Configures the optimizers.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            torch.optim.Optimizer: The optimizer.
        """
        if self.use_adam_optimizer:
            return torch.optim.Adam(self.parameters(), **self.optimizer_kwargs)

        return torch.optim.Adamax(self.parameters(), **self.optimizer_kwargs)


class Residual(nn.Module):
    """
    A residual layer that adds the output of a function to its input.

    Args:
        fn (nn.Module): The function to be applied to the input.

    References:
        https://github.com/bio-ontology-research-group/deepgo2/blob/main/deepgo/base.py#L6-L35
    """

    def __init__(self, fn):
        """
        Initialize the Residual layer with a given function.

        Args:
            fn (nn.Module): The function to be applied to the input.
        """
        super().__init__()
        self.fn = fn

    def forward(self, x):
        """
        Forward pass of the Residual layer.

        Args:
            x: Input tensor.

        Returns:
            torch.Tensor: The input tensor added to the result of applying the function `fn` to it.
        """
        return x + self.fn(x)


class MLPBlock(nn.Module):
    """
    A basic Multi-Layer Perceptron (MLP) block with one fully connected layer.

    Args:
        in_features (int): The number of input features.
        output_size (int): The number of output features.
        bias (boolean): Add bias to the linear layer
        layer_norm (boolean): Apply layer normalization
        dropout (float): The dropout value
        activation (nn.Module): The activation function to be applied after each fully connected layer.

    References:
        https://github.com/bio-ontology-research-group/deepgo2/blob/main/deepgo/base.py#L38-L73

    Example:
    ```python
    # Create an MLP block with 2 hidden layers and ReLU activation
    mlp_block = MLPBlock(input_size=64, output_size=10, activation=nn.ReLU())

    # Apply the MLP block to an input tensor
    input_tensor = torch.randn(32, 64)
    output = mlp_block(input_tensor)
    ```
    """

    def __init__(
        self,
        in_features,
        out_features,
        bias=True,
        layer_norm=True,
        dropout=0.1,
        activation=nn.ReLU,
    ):
        super().__init__()
        self.linear = nn.Linear(in_features, out_features, bias)
        self.activation = activation()
        self.layer_norm: Optional[nn.LayerNorm] = (
            nn.LayerNorm(out_features) if layer_norm else None
        )
        self.dropout: Optional[nn.Dropout] = nn.Dropout(dropout) if dropout else None

    def forward(self, x):
        x = self.activation(self.linear(x))
        if self.layer_norm:
            x = self.layer_norm(x)
        if self.dropout:
            x = self.dropout(x)
        return x



import abc
import multiprocessing as mp
from typing import Iterable

import torch
import tqdm

from chebai.models.base import ChebaiBaseNet

PROCESSORS = dict()


class ResultProcessor(abc.ABC):
    @classmethod
    def _identifier(cls) -> str:
        raise NotImplementedError

    def start(self):
        pass

    def close(self):
        pass

    def __init_subclass__(cls, **kwargs):
        assert (
            cls._identifier() not in PROCESSORS
        ), f"ResultProcessor {cls.__name__} does not have a unique identifier"
        PROCESSORS[cls._identifier()] = cls

    def process_prediction(self, proc_id, features, labels, pred, ident):
        raise NotImplementedError


class ResultFactory(abc.ABC):
    def __init__(
        self, model: ChebaiBaseNet, dataset, processors: Iterable[ResultProcessor]
    ):
        self._model = model
        self._reader = dataset.reader
        self.dataset = dataset
        self._processors = processors

    def _process_row(self, row):
        return row

    def _generate_predictions(self, data_path, raw=False, **kwargs):
        self._model.eval()
        collate = self._reader.COLLATOR()
        if raw:
            data_tuples = [
                (x["features"], x["ident"], self._reader.to_data(self._process_row(x)))
                for x in self.dataset._load_dict(data_path)
            ]
        else:
            data_tuples = [
                (x.get("raw_features", x["ident"]), x["ident"], x)
                for x in torch.load(data_path, weights_only=False)
            ]

        for raw_features, ident, row in tqdm.tqdm(data_tuples):
            raw_labels = row.get("labels")

            processable_data = self._model._process_batch(collate([row]), 0)

            model_output = self._model(processable_data)
            preds, labels = self._model._get_prediction_and_labels(
                processable_data, processable_data["labels"], model_output
            )
            d = dict(
                model_output=model_output,
                preds=preds,
                raw_features=raw_features,
                ident=ident,
                threshold=self._model.thres,
            )
            if raw_labels is not None:
                d["labels"] = raw_labels
            yield d

    def call_procs(self, args):
        proc_id, proc_args = args
        for proc in self._processors:
            try:
                proc.process_prediction(proc_id, **proc_args)
            except Exception:
                print("Could not process results for", proc_args["ident"])
                raise

    def execute(self, data_path, **kwargs):
        for proc in self._processors:
            proc.start()
        try:
            with mp.Pool():
                res = map(
                    self.call_procs,
                    enumerate(self._generate_predictions(data_path, **kwargs)),
                )
            for r in res:
                pass

        except:
            raise
        finally:
            for proc in self._processors:
                proc.close()



from os import makedirs
from tempfile import NamedTemporaryFile

import networkx as nx
import numpy as np
import torch
from matplotlib import cm
from matplotlib import pyplot as plt
from matplotlib import rc
from matplotlib.image import imread
from networkx.algorithms.isomorphism import GraphMatcher
from pysmiles.read_smiles import LOGGER, TokenType, _tokenize
from rdkit import Chem
from rdkit.Chem.Draw import MolToMPL  # , rdMolDraw2D
from rdkit.Chem.Draw import rdMolDraw2D

from chebai.preprocessing.datasets import JCI_500_COLUMNS_INT
from chebai.result.base import ResultProcessor


class AttentionMolPlot:
    def draw_attention_molecule(self, smiles, attention):
        pmol = self.read_smiles_with_index(smiles)
        rdmol = Chem.MolFromSmiles(smiles)
        if not rdmol:
            raise NoRDMolException
        rdmolx = self.mol_to_nx(rdmol)
        gm = GraphMatcher(pmol, rdmolx)
        iso = next(gm.isomorphisms_iter())
        token_to_node_map = {
            pmol.nodes[node]["token_index"]: iso[node] for node in pmol.nodes
        }
        d = rdMolDraw2D.MolDraw2DCairo(500, 500)
        cmap = cm.ScalarMappable(cmap=cm.Greens)

        aggr_attention_colors = cmap.to_rgba(
            np.max(attention[2:, :], axis=0), norm=False
        )
        cols = {
            token_to_node_map[token_index]: tuple(
                aggr_attention_colors[token_index].tolist()
            )
            for node, token_index in nx.get_node_attributes(pmol, "token_index").items()
        }
        highlight_atoms = [
            token_to_node_map[token_index]
            for node, token_index in nx.get_node_attributes(pmol, "token_index").items()
        ]
        rdMolDraw2D.PrepareAndDrawMolecule(
            d, rdmol, highlightAtoms=highlight_atoms, highlightAtomColors=cols
        )

        d.FinishDrawing()
        return d

    def plot_attentions(self, smiles, attention, threshold, labels):
        d = self.draw_attention_molecule(smiles, attention)
        cmap = cm.ScalarMappable(cmap=cm.Greens)
        attention_colors = cmap.to_rgba(attention, norm=False)
        num_tokens = sum(1 for _ in _tokenize(smiles))

        fig = plt.figure(figsize=(15, 15), facecolor="w")

        rc("font", **{"family": "monospace", "monospace": "DejaVu Sans Mono"})
        fig.tight_layout()

        ax2, ax = fig.subplots(2, 1, gridspec_kw={"height_ratios": [10, 1]})

        with NamedTemporaryFile(mode="wt", suffix=".png") as svg1:
            d.WriteDrawingText(svg1.name)
            ax2.imshow(imread(svg1.name))
        ax2.axis("off")
        ax2.spines["left"].set_position("center")
        ax2.spines["bottom"].set_position("zero")
        ax2.autoscale(tight=True)

        table = plt.table(
            cellText=[
                (["[CLS]"] + [t for _, t in _tokenize(smiles)])
                for _ in range(attention.shape[0])
            ],
            cellColours=attention_colors,
            cellLoc="center",
        )
        table.auto_set_column_width(list(range(num_tokens)))
        table.scale(1, 4)
        table.set_fontsize(26)

        ax.add_table(table)
        ax.axis("off")
        ax.spines["top"].set_position("zero")
        ax.autoscale(tight=True)

        self.counter += 1
        for w, label, predicted in labels:
            if predicted:
                cat = "p"
            else:
                cat = "n"
            if predicted == label:
                cat = "t" + cat
            else:
                cat = "f" + cat
            fig.savefig(
                f"/tmp/plots/{w}/{cat}_{self.counter}.png",
                transparent=False,
                bbox_inches="tight",
                pad_inches=0,
            )
        plt.close()

    @staticmethod
    def mol_to_nx(mol):
        G = nx.Graph()

        for atom in mol.GetAtoms():
            G.add_node(
                atom.GetIdx(),
                atomic_num=atom.GetAtomicNum(),
                formal_charge=atom.GetFormalCharge(),
                chiral_tag=atom.GetChiralTag(),
                hybridization=atom.GetHybridization(),
                num_explicit_hs=atom.GetNumExplicitHs(),
                is_aromatic=atom.GetIsAromatic(),
            )
        for bond in mol.GetBonds():
            G.add_edge(
                bond.GetBeginAtomIdx(),
                bond.GetEndAtomIdx(),
                bond_type=bond.GetBondType(),
            )
        return G

    @staticmethod
    def read_smiles_with_index(
        smiles,
        explicit_hydrogen=False,
        zero_order_bonds=True,
        reinterpret_aromatic=True,
    ):
        """
        This is just a re-implementation of pysmiles.read_smiles, that stores token indices
        """
        bond_to_order = {"-": 1, "=": 2, "#": 3, "$": 4, ":": 1.5, ".": 0}
        mol = nx.Graph()
        anchor = None
        idx = 0
        default_bond = 1
        next_bond = None
        branches = []
        ring_nums = {}
        for token_index, (tokentype, token) in enumerate(_tokenize(smiles)):
            if tokentype == TokenType.ATOM:
                mol.add_node(
                    idx,
                    token_index=token_index,
                    **parse_atom(token),  # noqa: F821
                )
                if anchor is not None:
                    if next_bond is None:
                        next_bond = default_bond
                    if next_bond or zero_order_bonds:
                        mol.add_edge(anchor, idx, order=next_bond)
                    next_bond = None
                anchor = idx
                idx += 1
            elif tokentype == TokenType.BRANCH_START:
                branches.append(anchor)
            elif tokentype == TokenType.BRANCH_END:
                anchor = branches.pop()
            elif tokentype == TokenType.BOND_TYPE:
                if next_bond is not None:
                    raise ValueError(
                        'Previous bond (order {}) not used. Overwritten by "{}"'.format(
                            next_bond, token
                        )
                    )
                next_bond = bond_to_order[token]
            elif tokentype == TokenType.RING_NUM:
                if token in ring_nums:
                    jdx, order = ring_nums[token]
                    if next_bond is None and order is None:
                        next_bond = default_bond
                    elif order is None:  # Note that the check is needed,
                        next_bond = next_bond  # But this could be pass.
                    elif next_bond is None:
                        next_bond = order
                    elif next_bond != order:  # Both are not None
                        raise ValueError(
                            "Conflicting bond orders for ring "
                            "between indices {}".format(token)
                        )
                    # idx is the index of the *next* atom we're adding. So: -1.
                    if mol.has_edge(idx - 1, jdx):
                        raise ValueError(
                            "Edge specified by marker {} already exists".format(token)
                        )
                    if idx - 1 == jdx:
                        raise ValueError(
                            "Marker {} specifies a bond between an "
                            "atom and itself".format(token)
                        )
                    if next_bond or zero_order_bonds:
                        mol.add_edge(idx - 1, jdx, order=next_bond)
                    next_bond = None
                    del ring_nums[token]
                else:
                    if idx == 0:
                        raise ValueError(
                            "Can't have a marker ({}) before an atom".format(token)
                        )
                    # idx is the index of the *next* atom we're adding. So: -1.
                    ring_nums[token] = (idx - 1, next_bond)
                    next_bond = None
            elif tokentype == TokenType.EZSTEREO:
                LOGGER.warning(
                    'E/Z stereochemical information, which is specified by "%s", will be discarded',
                    token,
                )
        if ring_nums:
            raise KeyError("Unmatched ring indices {}".format(list(ring_nums.keys())))

        # Time to deal with aromaticity. This is a mess, because it's not super
        # clear what aromaticity information has been provided, and what should be
        # inferred. In addition, to what extend do we want to provide a "sane"
        # molecule, even if this overrides what the SMILES string specifies?
        cycles = nx.cycle_basis(mol)
        ring_idxs = set()
        for cycle in cycles:
            ring_idxs.update(cycle)
        non_ring_idxs = set(mol.nodes) - ring_idxs
        for n_idx in non_ring_idxs:
            if mol.nodes[n_idx].get("aromatic", False):
                raise ValueError(
                    "You specified an aromatic atom outside of a"
                    " ring. This is impossible"
                )

        mark_aromatic_edges(mol)  # noqa: F405, F821
        fill_valence(mol)  # noqa: F405, F821
        if reinterpret_aromatic:
            mark_aromatic_atoms(mol)  # noqa: F405, F821
            mark_aromatic_edges(mol)  # noqa: F405, F821
            for idx, jdx in mol.edges:
                if (
                    not mol.nodes[idx].get("aromatic", False)
                    or not mol.nodes[jdx].get("aromatic", False)
                ) and mol.edges[idx, jdx].get("order", 1) == 1.5:
                    mol.edges[idx, jdx]["order"] = 1

        if explicit_hydrogen:
            add_explicit_hydrogens(mol)  # noqa: F405, F821
        else:
            remove_explicit_hydrogens(mol)  # noqa: F405, F821
        return mol


class AttentionOnMoleculesProcessor(AttentionMolPlot, ResultProcessor):
    def __init__(self, *args, headers=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.headers = headers

    def start(self):
        self.counter = 0

    @classmethod
    def _identifier(cls):
        return "platt"

    def filter(self, l_):
        return

    def process_prediction(
        self, proc_id, preds, raw_features, model_output, labels, **kwargs
    ):
        atts = torch.stack(model_output["attentions"]).squeeze(1).detach().numpy()
        predictions = preds.detach().numpy().squeeze(0) > 0.5
        if self.headers is None:
            headers = list(range(len(labels)))
        else:
            headers = self.headers

        for w in headers:
            makedirs(f"/tmp/plots/{w}", exist_ok=True)

        try:
            self.plot_attentions(
                raw_features,
                np.max(np.max(atts, axis=2), axis=1),
                0.4,
                [
                    (ident, label, predicted)
                    for label, ident, predicted in zip(labels, headers, predictions)
                    if (label or predicted)
                ],
            )
        except StopIteration:
            print("Could not match", raw_features)
        except NoRDMolException:
            pass


class LastLayerAttentionProcessor(AttentionMolPlot, ResultProcessor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def start(self):
        self.counter = 0
        for w in JCI_500_COLUMNS_INT:
            makedirs(f"/tmp/plots/{w}", exist_ok=True)

    @classmethod
    def _identifier(cls):
        return "platt_last"

    def filter(self, l_):
        return

    def process_prediction(self, raw_features, raw_labels, features, labels, pred):
        atts = torch.stack(pred["attentions"]).squeeze(1).detach().numpy()
        last_layer = np.max(atts, axis=2)[-1, :]
        if np.any(last_layer > 0.4):
            try:
                self.plot_attentions(
                    raw_features,
                    np.max(np.max(atts, axis=2), axis=1),
                    0.4,
                    [
                        ident
                        for present, ident in zip(labels, JCI_500_COLUMNS_INT)
                        if present
                    ],
                )
            except StopIteration:
                print("Could not match", raw_features)
            except NoRDMolException:
                pass


class SingletonAttentionProcessor(AttentionMolPlot, ResultProcessor):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def start(self):
        self.counter = 0
        for w in JCI_500_COLUMNS_INT:
            makedirs(f"/tmp/plots/{w}", exist_ok=True)

    @classmethod
    def _identifier(cls):
        return "platt_singles"

    def filter(self, l_):
        return

    def process_prediction(self, raw_features, raw_labels, features, labels, pred):
        atts = torch.stack(pred["attentions"]).squeeze(1).detach().numpy()
        if sum(labels) == 1:
            try:
                predictions = (
                    torch.sigmoid(pred["logits"]).detach().numpy().squeeze(0) > 0.5
                )
                self.plot_attentions(
                    raw_features,
                    np.max(np.average(atts, axis=2), axis=1),
                    0.4,
                    [
                        (ident, label, predicted)
                        for label, ident, predicted in zip(
                            labels, JCI_500_COLUMNS_INT, predictions
                        )
                        if (label or predicted)
                    ],
                )
            except StopIteration:
                print("Could not match", raw_features)
            except NoRDMolException:
                pass


class AttentionNetwork(ResultProcessor):
    def __init__(self, *args, headers=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.headers = headers
        self.i = 0

    @classmethod
    def _identifier(cls):
        return "platt_table"

    def start(self):
        self.counter = 0

    def process_prediction(
        self,
        proc_id,
        preds,
        raw_features,
        model_output,
        labels,
        ident=None,
        threshold=0.5,
        **kwargs,
    ):
        if self.headers is None:
            headers = list(range(len(labels)))
        else:
            headers = self.headers

        for w in headers:
            makedirs(f"plots/{w}", exist_ok=True)

        atts = torch.stack(model_output["attentions"]).squeeze(1).detach().numpy()
        predictions = preds.detach().numpy().squeeze(0) > 0.5
        plt.rcParams.update({"font.size": 8})
        try:
            attentions = atts
            tokens = ["[CLS]"] + [s for _, s in _tokenize(raw_features)]
            cmap = cm.ScalarMappable(cmap=cm.Greens)
            assert len(tokens) == attentions.shape[2]

            # rows = int((attentions.shape[1] + 2))
            width = len(tokens)
            # height = 12
            rdmol = Chem.MolFromSmiles(raw_features)
            if rdmol is not None:
                fig0 = MolToMPL(rdmol, fitImage=True)
                fig0.text(
                    0.1,
                    0,
                    "annotated:"
                    + ", ".join(
                        str(l_) for (l_, is_member) in zip(headers, labels) if is_member
                    )
                    + "\n"
                    + "predicted:"
                    + ", ".join(
                        str(l_)
                        for (l_, is_member) in zip(headers, predictions)
                        if is_member
                    ),
                    fontdict=dict(fontsize=10),
                )
                fig0.savefig(
                    f"plots/mol_{ident}.png",
                    bbox_inches="tight",
                    pad_inches=0,
                )
                plt.close(fig0)
                fig = plt.figure(figsize=(10 * 12, width // 3))
                l_tokens = {i: str(t) for i, t in enumerate(tokens)}
                r_tokens = {(len(tokens) + i): str(t) for i, t in enumerate(tokens)}
                labels = dict(list(l_tokens.items()) + list(r_tokens.items()))
                edges = [(l_, r) for r in r_tokens.keys() for l_ in l_tokens.keys()]
                g = nx.Graph()
                g.add_nodes_from(l_tokens, bipartite=0)
                g.add_nodes_from(r_tokens, bipartite=1)
                g.add_edges_from(edges)
                pos = np.array(
                    [(0, -i) for i in range(len(l_tokens))]
                    + [(1, -i) for i in range(len(l_tokens))]
                )

                offset = np.array(
                    [(1, 0) for i in range(len(l_tokens))]
                    + [(1, 0) for i in range(len(l_tokens))]
                )
                # axes = fig.subplots(1, 6 * 8 + 5, subplot_kw=dict(frameon=False))

                ax = fig.add_subplot(111)
                ax.axis("off")
                for layer in range(attentions.shape[0]):
                    for head in range(attentions.shape[1]):
                        index = 8 * (layer) + head + layer + 1

                        at = np.concatenate([a for a in attentions[layer, head]])
                        col = cmap.cmap(at)
                        col[:, 3] = at
                        nx.draw_networkx(
                            g,
                            pos=pos + (index * offset),
                            edge_color=col,
                            ax=ax,
                            labels=labels,
                            node_color="none",
                            node_size=8,
                        )
                        # sns.heatmap(attentions[i,j], linewidth=0.5, ax=ax, cmap=cm.Greens, square=True, vmin=0, vmax=1, xticklabels=tokens, yticklabels=tokens)
                fig.subplots_adjust()
                fig.savefig(
                    f"plots/att_{ident}.png",
                    # transparent=True,
                    bbox_inches="tight",
                    pad_inches=0,
                    dpi=100,
                )

            plt.close()
        except StopIteration:
            print("Could not match", raw_features)
        except NoRDMolException:
            pass
        finally:
            plt.close()


class NoRDMolException(Exception):
    pass



from urllib import request
import csv
import os
import shutil
from typing import Dict, List

from sklearn.model_selection import train_test_split
import torch

from chebai.preprocessing import reader as dr
from chebai.preprocessing.datasets.base import XYBaseDataModule


class Lipo(XYBaseDataModule):
    HEADERS = [
        "exp",
    ]

    @property
    def _name(self):
        return "Lipo"

    @property
    def label_number(self):
        return 1

    @property
    def raw_file_names(self):
        return ["Lipo.csv"]

    # @property
    # def processed_file_names(self):
    #     return ["test.pt", "train.pt", "validation.pt"]

    @property
    def processed_file_names_dict(self) -> dict:
        return {
            "test": "test.pt",
            "train": "train.pt",
            "validation": "validation.pt",
        }

    def download(self):
        # download
        with open(os.path.join(self.raw_dir, "Lipo.csv"), "ab") as dst:
            with request.urlopen(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/Lipophilicity.csv",
            ) as src:
                shutil.copyfileobj(src, dst)

    def setup_processed(self):
        print("Create splits")
        data = list(self._load_data_from_file(os.path.join(self.raw_dir, "Lipo.csv")))
        print(len(data))

        train_split, test_split = train_test_split(
            data, test_size=self.test_split, shuffle=True
        )
        train_split, validation_split = train_test_split(
            train_split, test_size=self.validation_split, shuffle=True
        )

        if False:
            train_split, test_split = train_test_split(
                data, train_size=self.train_split, shuffle=True
            )
            test_split, validation_split = train_test_split(
                test_split, train_size=0.5, shuffle=True
            )
        for k, split in [
            ("test", test_split),
            ("train", train_split),
            ("validation", validation_split),
        ]:
            print("transform", k)
            torch.save(
                split,
                os.path.join(self.processed_dir, f"{k}.pt"),
            )

    def setup(self, **kwargs):
        if any(
            not os.path.isfile(os.path.join(self.raw_dir, f))
            for f in self.raw_file_names
        ):
            self.download()
        print(
            [
                not os.path.isfile(os.path.join(self.processed_dir, f))
                for f in self.processed_file_names
            ]
        )
        if any(
            not os.path.isfile(os.path.join(self.processed_dir, f))
            for f in self.processed_file_names
        ):
            self.setup_processed()

        self._after_setup()

    def _set_processed_data_props(self):
        """
        Load processed data and extract metadata.

        Sets:
            - self._num_of_labels: Number of target labels in the dataset.
            - self._feature_vector_size: Maximum feature vector length across all data points.
        """
        pt_file_path = os.path.join(
            self.processed_dir, self.processed_file_names_dict["train"]
        )
        data_pt = torch.load(pt_file_path, weights_only=False)

        self._num_of_labels = len(data_pt[0]["labels"])
        self._feature_vector_size = max(len(d["features"]) for d in data_pt)

    def _load_dict(self, input_file_path: str) -> List[Dict]:
        """Loads data from a CSV file.

        Args:
            input_file_path (str): Path to the CSV file.

        Returns:
            List[Dict]: List of data dictionaries.
        """
        smiles_l = []
        labels_l = []
        with open(input_file_path, "r") as input_file:
            reader = csv.DictReader(input_file)
            print(reader.fieldnames)
            for row in reader:
                smiles_l.append(row["smiles"])
                labels_l.append(float(row["exp"]))

        for i in range(0, len(smiles_l)):
            yield dict(features=smiles_l[i], labels=[labels_l[i]], ident=i)
            # yield self.reader.to_data(dict(features=smiles_l[i], labels=[labels_l[i]], ident=i))

    def _perform_data_preparation(self, *args, **kwargs) -> None:
        pass


class LipoChem(Lipo):
    """Chemical data reader for the solubility dataset."""

    READER = dr.ChemDataReader


class FreeSolv(XYBaseDataModule):
    HEADERS = [
        "expt",
    ]

    @property
    def _name(self):
        return "FreeSolv"

    @property
    def label_number(self):
        return 1

    @property
    def raw_file_names(self):
        return ["FreeSolv.csv"]

    # @property
    # def processed_file_names(self):
    #     return ["test.pt", "train.pt", "validation.pt"]

    @property
    def processed_file_names_dict(self) -> dict:
        return {
            "test": "test.pt",
            "train": "train.pt",
            "validation": "validation.pt",
        }

    def download(self):
        # download
        with open(os.path.join(self.raw_dir, "FreeSolv.csv"), "ab") as dst:
            with request.urlopen(
                "https://deepchemdata.s3-us-west-1.amazonaws.com/datasets/SAMPL.csv",
            ) as src:
                shutil.copyfileobj(src, dst)

    def setup_processed(self):
        print("Create splits")
        data = list(
            self._load_data_from_file(os.path.join(self.raw_dir, "FreeSolv.csv"))
        )
        print(len(data))
        train_split, test_split = train_test_split(
            data, test_size=self.test_split, shuffle=True
        )
        train_split, validation_split = train_test_split(
            train_split, test_size=self.validation_split, shuffle=True
        )

        if False:
            train_split, test_split = train_test_split(
                data, train_size=self.train_split, shuffle=True
            )
            test_split, validation_split = train_test_split(
                test_split, train_size=0.5, shuffle=True
            )
        for k, split in [
            ("test", test_split),
            ("train", train_split),
            ("validation", validation_split),
        ]:
            print("transform", k)
            torch.save(
                split,
                os.path.join(self.processed_dir, f"{k}.pt"),
            )

    def setup(self, **kwargs):
        if any(
            not os.path.isfile(os.path.join(self.raw_dir, f))
            for f in self.raw_file_names
        ):
            self.download()
        print(
            [
                not os.path.isfile(os.path.join(self.processed_dir, f))
                for f in self.processed_file_names
            ]
        )
        if any(
            not os.path.isfile(os.path.join(self.processed_dir, f))
            for f in self.processed_file_names
        ):
            self.setup_processed()

        self._after_setup()

    def _set_processed_data_props(self):
        """
        Load processed data and extract metadata.

        Sets:
            - self._num_of_labels: Number of target labels in the dataset.
            - self._feature_vector_size: Maximum feature vector length across all data points.
        """
        pt_file_path = os.path.join(
            self.processed_dir, self.processed_file_names_dict["train"]
        )
        data_pt = torch.load(pt_file_path, weights_only=False)

        self._num_of_labels = len(data_pt[0]["labels"])
        self._feature_vector_size = max(len(d["features"]) for d in data_pt)

    def _load_dict(self, input_file_path: str) -> List[Dict]:
        """Loads data from a CSV file.

        Args:
            input_file_path (str): Path to the CSV file.

        Returns:
            List[Dict]: List of data dictionaries.
        """
        smiles_l = []
        labels_l = []
        with open(input_file_path, "r") as input_file:
            reader = csv.DictReader(input_file)
            print(reader.fieldnames)
            for row in reader:
                smiles_l.append(row["smiles"])
                labels_l.append(float(row["expt"]))

        for i in range(0, len(smiles_l)):
            yield dict(features=smiles_l[i], labels=[labels_l[i]], ident=i)
            # yield self.reader.to_data(dict(features=smiles_l[i], labels=[labels_l[i]], ident=i))

    def _perform_data_preparation(self, *args, **kwargs) -> None:
        pass


class FreeSolvChem(FreeSolv):
    """Chemical data reader for the solubility dataset."""

    READER = dr.ChemDataReader



import os
import random
from abc import ABC, abstractmethod
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Generator, List, Optional, Tuple, Union

import lightning as pl
import numpy as np
import pandas as pd
import torch
import tqdm
from lightning.pytorch.core.datamodule import LightningDataModule
from lightning_utilities.core.rank_zero import rank_zero_info
from torch.utils.data import DataLoader

from chebai.preprocessing import reader as dr

if TYPE_CHECKING:
    import networkx as nx


class XYBaseDataModule(LightningDataModule):
    """
    Base class for data modules.

    This class provides a base implementation for loading and preprocessing datasets.
    It inherits from `LightningDataModule` and defines common properties and methods for data loading and processing.

    Args:
        batch_size (int): The batch size for data loading. Default is 1.
        test_split (float): The ratio of test data to total data. Default is 0.1.
        validation_split (float): The ratio of validation data to total data. Default is 0.05.
        reader_kwargs (dict): Additional keyword arguments to be passed to the data reader. Default is None.
        prediction_kind (str): The kind of prediction to be performed (only relevant for the predict_dataloader). Default is "test".
        data_limit (Optional[int]): The maximum number of data samples to load. If set to None, the complete dataset will be used. Default is None.
        label_filter (Optional[int]): The index of the label to filter. Default is None.
        balance_after_filter (Optional[float]): The ratio of negative samples to positive samples after filtering. Default is None.
        num_workers (int): The number of worker processes for data loading. Default is 1.
        chebi_version (int): The version of ChEBI to use. Default is 200.
        inner_k_folds (int): The number of folds for inner cross-validation. Use -1 to disable inner cross-validation. Default is -1.
        fold_index (Optional[int]): The index of the fold to use for training and validation. Default is None.
        base_dir (Optional[str]): The base directory for storing processed and raw data. Default is None.
        **kwargs: Additional keyword arguments.

    Attributes:
        READER (DataReader): The data reader class to use.
        reader (DataReader): An instance of the data reader class.
        test_split (float): The ratio of test data to total data.
        validation_split (float): The ratio of validation data to total data.
        batch_size (int): The batch size for data loading.
        prediction_kind (str): The kind of prediction to be performed.
        data_limit (Optional[int]): The maximum number of data samples to load.
        label_filter (Optional[int]): The index of the label to filter.
        balance_after_filter (Optional[float]): The ratio of negative samples to positive samples after filtering.
        num_workers (int): The number of worker processes for data loading.
        chebi_version (int): The version of ChEBI to use.
        inner_k_folds (int): The number of folds for inner cross-validation. If it is less than to, no cross-validation will be performed.
        fold_index (Optional[int]): The index of the fold to use for training and validation (only relevant for cross-validation).
        _base_dir (Optional[str]): The base directory for storing processed and raw data.
        raw_dir (str): The directory for storing raw data.
        processed_dir (str): The directory for storing processed data.
        fold_dir (str): The name of the directory where the folds from inner cross-validation are stored.
        _name (str): The name of the data module.

    """

    READER = dr.DataReader

    def __init__(
        self,
        batch_size: int = 1,
        test_split: Optional[float] = 0.1,
        validation_split: Optional[float] = 0.05,
        reader_kwargs: Optional[dict] = None,
        prediction_kind: str = "test",
        data_limit: Optional[int] = None,
        label_filter: Optional[int] = None,
        balance_after_filter: Optional[float] = None,
        num_workers: int = 1,
        persistent_workers: bool = True,
        chebi_version: int = 200,
        inner_k_folds: int = -1,  # use inner cross-validation if > 1
        fold_index: Optional[int] = None,
        base_dir: Optional[str] = None,
        n_token_limit: Optional[int] = None,
        **kwargs,
    ):
        super().__init__()
        if reader_kwargs is None:
            reader_kwargs = dict()
        self.reader = self.READER(**reader_kwargs)
        self.test_split = test_split
        self.validation_split = validation_split

        self.batch_size = batch_size
        self.prediction_kind = prediction_kind
        self.data_limit = data_limit
        self.label_filter = label_filter
        assert (balance_after_filter is not None) or (
            self.label_filter is None
        ), "Filter balancing requires a filter"
        self.balance_after_filter = balance_after_filter
        self.num_workers = num_workers
        self.persistent_workers: bool = bool(persistent_workers)
        self.chebi_version = chebi_version
        assert type(inner_k_folds) is int
        self.inner_k_folds = inner_k_folds
        self.use_inner_cross_validation = (
            inner_k_folds > 1
        )  # only use cv if there are at least 2 folds
        assert (
            fold_index is None or self.use_inner_cross_validation is not None
        ), "fold_index can only be set if cross validation is used"
        if fold_index is not None and self.inner_k_folds is not None:
            assert (
                fold_index < self.inner_k_folds
            ), "fold_index can't be larger than the total number of folds"
        self.fold_index = fold_index
        self._base_dir = base_dir
        self.n_token_limit = n_token_limit
        os.makedirs(self.raw_dir, exist_ok=True)
        os.makedirs(self.processed_dir, exist_ok=True)
        if self.use_inner_cross_validation:
            os.makedirs(os.path.join(self.raw_dir, self.fold_dir), exist_ok=True)
            os.makedirs(os.path.join(self.processed_dir, self.fold_dir), exist_ok=True)
        self.save_hyperparameters()

        self._num_of_labels = None
        self._feature_vector_size = None
        self._prepare_data_flag = 1
        self._setup_data_flag = 1

    @property
    def num_of_labels(self):
        assert self._num_of_labels is not None, "num of labels must be set"
        return self._num_of_labels

    @property
    def feature_vector_size(self):
        assert (
            self._feature_vector_size is not None
        ), "size of feature vector must be set"
        return self._feature_vector_size

    @property
    def identifier(self) -> tuple:
        """Identifier for the dataset."""
        return (self.reader.name(),)

    @property
    def full_identifier(self) -> tuple:
        """Full identifier for the dataset."""
        return (self._name, *self.identifier)

    @property
    def base_dir(self) -> str:
        """Common base directory for processed and raw directories."""
        if self._base_dir is not None:
            return self._base_dir
        return os.path.join("data", self._name)

    @property
    def processed_dir_main(self) -> str:
        """Name of the directory where processed (but not tokenized) data is stored."""
        return os.path.join(self.base_dir, "processed")

    @property
    def processed_dir(self) -> str:
        """Name of the directory where the processed and tokenized data is stored."""
        return os.path.join(self.processed_dir_main, *self.identifier)

    @property
    def raw_dir(self) -> str:
        """Name of the directory where the raw data is stored."""
        return os.path.join(self.base_dir, "raw")

    @property
    def fold_dir(self) -> str:
        """Name of the directory where the folds from inner cross-validation (i.e., the train and val sets) are stored."""
        return f"cv_{self.inner_k_folds}_fold"

    @property
    @abstractmethod
    def _name(self) -> str:
        """
        Abstract property representing the name of the data module.

        This property should be implemented in subclasses to provide a unique name for the data module.
        The name is used to create subdirectories within the base directory or `processed_dir_main`
        for storing relevant data associated with this module.

        Returns:
            str: The name of the data module.
        """
        pass

    def _filter_labels(self, row: dict) -> dict:
        """
        Filter labels based on `label_filter`.
        This method selects specific labels from the `labels` list within the row dictionary
        according to the index or indices provided by the `label_filter` attribute of the class.

        Args:
            row (dict): A dictionary containing the row data.

        Returns:
            dict: The filtered row data.
        """
        row["labels"] = [row["labels"][self.label_filter]]
        return row

    def load_processed_data(
        self, kind: Optional[str] = None, filename: Optional[str] = None
    ) -> List:
        """
        Load processed data from a file. Either the kind or the filename has to be provided. If both are provided, the
        filename is used.

        Args:
            kind (str, optional): The kind of dataset to load such as "train", "val" or "test". Defaults to None.
            filename (str, optional): The name of the file to load the dataset from. Defaults to None.

        Returns:
            List: The loaded processed data.

        Raises:
            ValueError: If both kind and filename are None.
        """
        if kind is None and filename is None:
            raise ValueError(
                "Either kind or filename is required to load the correct dataset, both are None"
            )
        # if both kind and filename are given, use filename
        if kind is not None and filename is None:
            try:
                # processed_file_names_dict is only implemented for _ChEBIDataExtractor
                if self.use_inner_cross_validation and kind != "test":
                    filename = self.processed_file_names_dict[
                        f"fold_{self.fold_index}_{kind}"
                    ]
                else:
                    filename = self.processed_file_names_dict[kind]
            except NotImplementedError:
                filename = f"{kind}.pt"
        return torch.load(
            os.path.join(self.processed_dir, filename), weights_only=False
        )

    def dataloader(self, kind: str, **kwargs) -> DataLoader:
        """
        Returns a DataLoader object for the specified kind (train, val or test) of data.

        Args:
            kind (str): The kind indicates whether it is a train, val or test data to load.
            **kwargs: Additional keyword arguments.

        Returns:
            DataLoader: A DataLoader object.
        """
        rank_zero_info(
            f"Loading {kind} data... (datamodule.current_epoch={self.curr_epoch if hasattr(self, 'curr_epoch') else 'N/A'})"
        )
        dataset = self.load_processed_data(kind)
        if "ids" in kwargs:
            ids = kwargs.pop("ids")
            _dataset = []
            for i in range(len(dataset)):
                if i in ids:
                    _dataset.append(dataset[i])
            dataset = _dataset
        if self.label_filter is not None:
            original_len = len(dataset)
            dataset = [self._filter_labels(r) for r in dataset]
            positives = [r for r in dataset if r["labels"][0]]
            negatives = [r for r in dataset if not r["labels"][0]]
            if self.balance_after_filter is not None:
                negative_length = min(
                    original_len, int(len(positives) * self.balance_after_filter)
                )
                dataset = positives + negatives[:negative_length]
            else:
                dataset = positives + negatives
            random.shuffle(dataset)
        if self.data_limit is not None:
            dataset = dataset[: self.data_limit]
        return DataLoader(
            dataset,
            collate_fn=self.reader.collator,
            batch_size=self.batch_size,
            **kwargs,
        )

    @staticmethod
    def _load_dict(
        input_file_path: str,
    ) -> Generator[Dict[str, Any], None, None]:
        """
        Load data from a file and return a dictionary.

        Args:
            input_file_path (str): The path to the input file.

        Yields:
            dict: A dictionary containing the features and labels.
        """
        with open(input_file_path, "r") as input_file:
            for row in input_file:
                smiles, labels = row.split("\t")
                yield dict(features=smiles, labels=labels)

    @staticmethod
    def _get_data_size(input_file_path: str) -> int:
        """
        Get the number of lines in a file.

        Args:
            input_file_path (str): The path to the input file.

        Returns:
            int: The number of lines in the file.
        """
        with open(input_file_path, "r") as f:
            return sum(1 for _ in f)

    def _load_data_from_file(self, path: str) -> List[Dict[str, Any]]:
        """
        Load data from a file and return a list of dictionaries.

        Args:
            path (str): The path to the input file.

        Returns:
            List: A list of dictionaries containing the features and labels.
        """

        lines = self._get_data_size(path)
        print(f"Processing {lines} lines...")
        data = [
            self.reader.to_data(d)
            for d in tqdm.tqdm(self._load_dict(path), total=lines)
            if d["features"] is not None
        ]
        # filter for missing features in resulting data, keep features length below token limit
        data = [
            val
            for val in data
            if val["features"] is not None
            and (
                self.n_token_limit is None or len(val["features"]) <= self.n_token_limit
            )
        ]

        return data

    def train_dataloader(self, *args, **kwargs) -> Union[DataLoader, List[DataLoader]]:
        """
        Returns the train DataLoader.

        Args:
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            DataLoader: A DataLoader object for training data.
        """
        return self.dataloader(
            "train",
            shuffle=True,
            num_workers=self.num_workers,
            persistent_workers=self.persistent_workers,
            **kwargs,
        )

    def val_dataloader(self, *args, **kwargs) -> Union[DataLoader, List[DataLoader]]:
        """
        Returns the validation DataLoader.

        Args:
            *args: Additional positional arguments (unused).
            **kwargs: Additional keyword arguments, passed to dataloader().

        Returns:
            Union[DataLoader, List[DataLoader]]: A DataLoader object for validation data.
        """
        return self.dataloader(
            "validation",
            shuffle=False,
            num_workers=self.num_workers,
            persistent_workers=self.persistent_workers,
            **kwargs,
        )

    def test_dataloader(self, *args, **kwargs) -> Union[DataLoader, List[DataLoader]]:
        """
        Returns the test DataLoader.

        Args:
            *args: Additional positional arguments (unused).
            **kwargs: Additional keyword arguments, passed to dataloader().

        Returns:
            Union[DataLoader, List[DataLoader]]: A DataLoader object for test data.
        """
        return self.dataloader("test", shuffle=False, **kwargs)

    def predict_dataloader(
        self, *args, **kwargs
    ) -> Union[DataLoader, List[DataLoader]]:
        """
        Returns the predict DataLoader.

        Args:
            *args: Additional positional arguments (unused).
            **kwargs: Additional keyword arguments, passed to dataloader().

        Returns:
            Union[DataLoader, List[DataLoader]]: A DataLoader object for prediction data.
        """
        return self.dataloader(self.prediction_kind, shuffle=False, **kwargs)

    def prepare_data(self, *args, **kwargs) -> None:
        if self._prepare_data_flag != 1:
            return

        self._prepare_data_flag += 1
        self._perform_data_preparation(*args, **kwargs)
        self._after_prepare_data(*args, **kwargs)

    def _perform_data_preparation(self, *args, **kwargs) -> None:
        raise NotImplementedError

    def _after_prepare_data(self, *args, **kwargs) -> None:
        """
        Hook to perform additional pre-processing after pre-processed data is available.
        """
        ...

    def setup(self, *args, **kwargs) -> None:
        """
        Setup the data module.

        This method checks for the processed data and sets up the data module for training, validation, and testing.

        Args:
            **kwargs: Additional keyword arguments.
        """
        if self._setup_data_flag != 1:
            return

        self._setup_data_flag += 1

        rank_zero_info(f"Check for processed data in {self.processed_dir}")
        rank_zero_info(f"Cross-validation enabled: {self.use_inner_cross_validation}")
        rank_zero_info(f"Looking for files: {self.processed_file_names}")
        if any(
            not os.path.isfile(os.path.join(self.processed_dir, f))
            for f in self.processed_file_names
        ):
            self.setup_processed()

        self._after_setup(**kwargs)

    def _after_setup(self, **kwargs):
        """
        Finalize the setup process after ensuring the processed data is available.

        This method performs post-setup tasks like finalizing the reader and setting internal properties.
        """
        self.reader.on_finish()
        self._set_processed_data_props()

    def _set_processed_data_props(self):
        """
        Load processed data and extract metadata.

        Sets:
            - self._num_of_labels: Number of target labels in the dataset.
            - self._feature_vector_size: Maximum feature vector length across all data points.
        """
        pt_file_path = os.path.join(
            self.processed_dir, self.processed_file_names_dict["data"]
        )
        data_pt = torch.load(pt_file_path, weights_only=False)

        self._num_of_labels = len(data_pt[0]["labels"])
        self._feature_vector_size = max(len(d["features"]) for d in data_pt)

        print(
            f"Number of samples in encoded data ({pt_file_path}): {len(data_pt)} samples"
        )
        print(f"Number of labels for loaded data: {self._num_of_labels}")
        print(f"Feature vector size: {self._feature_vector_size}")

    def setup_processed(self):
        """
        Setup the processed data.

        This method should be implemented by subclasses to handle the specific setup of processed data.
        """
        raise NotImplementedError

    @property
    def processed_main_file_names_dict(self) -> dict:
        """
        Returns a dictionary mapping processed data file names.

        Returns:
            dict: A dictionary mapping dataset key to their respective file names.
                  For example, {"data": "data.pkl"}.
        """
        raise NotImplementedError

    @property
    def processed_main_file_names(self) -> List[str]:
        """
        Returns a list of file names for processed data (before tokenization).

        Returns:
            List[str]: A list of file names corresponding to the processed data.
        """
        return list(self.processed_main_file_names_dict.values())

    @property
    def processed_file_names_dict(self) -> dict:
        """
        Returns a dictionary for the processed and tokenized data files.

        Returns:
            dict: A dictionary mapping dataset keys to their respective file names.
                  For example, {"data": "data.pt"}.
        """
        raise NotImplementedError

    @property
    def processed_file_names(self) -> List[str]:
        """
        Returns a list of file names for processed data.

        Returns:
            List[str]: A list of file names corresponding to the processed data.
        """
        return list(self.processed_file_names_dict.values())

    @property
    def raw_file_names(self) -> List[str]:
        """
        Returns the list of raw file names.

        Returns:
            List[str]: The list of raw file names.
        """
        return list(self.raw_file_names_dict.values())

    @property
    def raw_file_names_dict(self) -> dict:
        """
        Returns the dictionary of raw file names (i.e., files that are directly obtained from an external source).

        This property should be implemented by subclasses to provide the dictionary of raw file names.

        Returns:
            dict: The dictionary of raw file names.
        """
        raise NotImplementedError


class MergedDataset(XYBaseDataModule):
    MERGED = []

    @property
    def _name(self) -> str:
        """
        Returns a concatenated name of all subset names.
        """
        return "+".join(s._name for s in self.subsets)

    def __init__(
        self,
        batch_size: int = 1,
        train_split: float = 0.85,
        reader_kwargs: Union[None, List[dict]] = None,
        **kwargs,
    ):
        """
        Args:
            batch_size (int): Batch size for data loaders.
            train_split (float): Fraction of data to use for training.
            reader_kwargs (Union[None, List[dict]]): Optional arguments for subset readers.
            **kwargs: Additional arguments to pass to LightningDataModule.
        """
        if reader_kwargs is None:
            reader_kwargs = [None for _ in self.MERGED]
        self.train_split = train_split
        self.batch_size = batch_size
        self.subsets = [
            s(train_split=train_split, reader_kwargs=kws)
            for s, kws in zip(self.MERGED, reader_kwargs)
        ]
        self.reader = self.subsets[0].reader
        os.makedirs(self.processed_dir, exist_ok=True)
        super(pl.LightningDataModule, self).__init__(**kwargs)

    def _perform_data_preparation(self):
        """
        Placeholder for data preparation logic.
        """
        for s in self.subsets:
            s.prepare_data()

    def setup(self, **kwargs):
        """
        Setup the data module.

        This method checks for the processed data and sets up the data module for training, validation, and testing.

        Args:
            **kwargs: Additional keyword arguments.
        """
        if self._setup_data_flag != 1:
            return

        self._setup_data_flag += 1
        for s in self.subsets:
            s.setup(**kwargs)

        self._set_processed_data_props()

    def dataloader(self, kind: str, **kwargs) -> DataLoader:
        """
        Creates a DataLoader for a specific subset.

        Args:
            kind (str): Kind of data loader ('train', 'validation', or 'test').
            **kwargs: Additional arguments passed to DataLoader.

        Returns:
            DataLoader: DataLoader object for the specified subset.
        """
        subdatasets = [
            torch.load(os.path.join(s.processed_dir, f"{kind}.pt"), weights_only=False)
            for s in self.subsets
        ]
        dataset = [
            self._process_data(i, d)
            for i, (s, lim) in enumerate(zip(subdatasets, self.limits))
            for d in (s if lim is None else s[:lim])
        ]
        return DataLoader(
            dataset,
            collate_fn=self.reader.collator,
            batch_size=self.batch_size,
            **kwargs,
        )

    def train_dataloader(self, *args, **kwargs) -> Union[DataLoader, List[DataLoader]]:
        """
        Returns the training DataLoader.
        """
        return self.dataloader("train", shuffle=True, **kwargs)

    def val_dataloader(self, *args, **kwargs) -> Union[DataLoader, List[DataLoader]]:
        """
        Returns the validation DataLoader.
        """
        return self.dataloader("validation", shuffle=False, **kwargs)

    def test_dataloader(self, *args, **kwargs) -> Union[DataLoader, List[DataLoader]]:
        """
        Returns the test DataLoader.
        """
        return self.dataloader("test", shuffle=False, **kwargs)

    def _process_data(self, subset_id: int, data: dict) -> dict:
        """
        Processes data from a subset.

        Args:
            subset_id (int): Index of the subset.
            data (dict): Data from the subset.

        Returns:
            dict: Processed data with 'features', 'labels', and 'ident' keys.
        """
        return dict(
            features=data["features"], labels=data["labels"], ident=data["ident"]
        )

    def setup_processed(self):
        """
        Placeholder for setup logic after data processing.
        """
        pass

    @property
    def processed_file_names(self) -> List[str]:
        """
        Returns the list of processed file names.
        """
        return ["test.pt", "train.pt", "validation.pt"]

    @property
    def limits(self):
        """
        Returns None, assuming no limits on data slicing.
        """
        return None


class _DynamicDataset(XYBaseDataModule, ABC):
    """
    A class for extracting and processing data from the given dataset.

    The processed and transformed data is stored in `data.pkl` and `data.pt` format as a whole respectively,
    rather than as separate  train, validation, and test splits, with dynamic splitting of data.pt occurring at runtime.
    The `_DynamicDataset` class manages data splits by either generating them during execution or retrieving them from
    a CSV file.
    If no split file path is provided, `_generate_dynamic_splits` creates the training, validation, and test splits
    from the encoded/transformed data, storing them in `_dynamic_df_train`, `_dynamic_df_val`, and `_dynamic_df_test`.
    When a split file path is provided, `_retrieve_splits_from_csv` loads splits from the CSV file, which must
    include 'id' and 'split' columns.
    The `dynamic_split_dfs` property ensures that the necessary splits are loaded as required.

    Args:
        dynamic_data_split_seed (int, optional): The seed for random data splitting. Defaults to 42.
        splits_file_path (str, optional): Path to the splits CSV file. Defaults to None.
        apply_label_filter (Optional[str]): Path to a classes.txt file - only labels that are in the labels filter
        file will be used (in that order). All labels in the label filter have to be present in the dataset. This filter
        is only active when loading splits from a CSV file. Defaults to None.
        apply_id_filter (Optional[str]): Path to a data.pt file from a different dataset - only IDs that are in the
        id filter file will be used. Defaults to None. This filter is only active when loading splits from a CSV file.
        **kwargs: Additional keyword arguments passed to XYBaseDataModule.

    Attributes:
        dynamic_data_split_seed (int): The seed for random data splitting, default is 42.
        splits_file_path (Optional[str]): Path to the CSV file containing split assignments.
        apply_label_filter (Optional[str]): Path to a classes.txt file for label filtering.
        apply_id_filter (Optional[str]): Path to a data.pt file for ID filtering.
    """

    # ---- Index for columns of processed `data.pkl` (should be derived from `_graph_to_raw_dataset` method) ------
    _ID_IDX: int = None
    _DATA_REPRESENTATION_IDX: int = None
    _LABELS_START_IDX: int = None

    def __init__(
        self,
        apply_label_filter: Optional[str] = None,
        apply_id_filter: Optional[str] = None,
        **kwargs,
    ):
        super(_DynamicDataset, self).__init__(**kwargs)
        self.dynamic_data_split_seed = int(kwargs.get("seed", 42))  # default is 42
        # Class variables to store the dynamics splits
        self._dynamic_df_train = None
        self._dynamic_df_test = None
        self._dynamic_df_val = None
        # Path of csv file which contains a list of ids & their assignment to a dataset (either train,
        # validation or test).
        self.splits_file_path = self._validate_splits_file_path(
            kwargs.get("splits_file_path", None)
        )
        self.apply_label_filter = apply_label_filter
        self.apply_id_filter = apply_id_filter
        self._data_pkl_filename: str = "data.pkl"

    @staticmethod
    def _validate_splits_file_path(splits_file_path: Optional[str]) -> Optional[str]:
        """
        Validates the file in provided splits file path.

        Args:
            splits_file_path (Optional[str]): Path to the splits CSV file.

        Returns:
            Optional[str]: Validated splits file path if checks pass, None if splits_file_path is None.

        Raises:
            FileNotFoundError: If the splits file does not exist.
            ValueError: If splits file is empty or missing required columns ('id' and/or 'split'), or not a CSV file.
        """
        if splits_file_path is None:
            return None

        if not os.path.isfile(splits_file_path):
            raise FileNotFoundError(f"File {splits_file_path} does not exist")

        file_size = os.path.getsize(splits_file_path)
        if file_size == 0:
            raise ValueError(f"File {splits_file_path} is empty")

        # Check if the file has a CSV extension
        if not splits_file_path.lower().endswith(".csv"):
            raise ValueError(f"File {splits_file_path} is not a CSV file")

        # Read the first row of CSV file into a DataFrame
        splits_df = pd.read_csv(splits_file_path, nrows=1)

        # Check if 'id' and 'split' columns are in the DataFrame
        required_columns = {"id", "split"}
        if not required_columns.issubset(splits_df.columns):
            raise ValueError(
                f"CSV file {splits_file_path} is missing required columns ('id' and/or 'split')."
            )

        return splits_file_path

    # ------------------------------ Phase: Prepare data -----------------------------------
    def _perform_data_preparation(self, *args: Any, **kwargs: Any) -> None:
        """
        Prepares the data for the dataset.

        This method checks for the presence of raw data in the specified directory.
        If the raw data is missing, it fetches the ontology and creates a dataframe and saves it to a data.pkl file.

        The resulting dataframe/pickle file is expected to contain columns with the following structure:
            - Column at index `self._ID_IDX`: ID of data instance
            - Column at index `self._DATA_REPRESENTATION_IDX`: Sequence representation of the protein
            - Column from index `self._LABELS_START_IDX` onwards: Labels

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            None
        """
        print("Checking for processed data in", self.processed_dir_main)

        processed_name = self.processed_main_file_names_dict["data"]
        if not os.path.isfile(os.path.join(self.processed_dir_main, processed_name)):
            print(f"Missing processed data file (`{processed_name}` file)")
            os.makedirs(self.processed_dir_main, exist_ok=True)
            data_path = self._download_required_data()
            g = self._extract_class_hierarchy(data_path)
            data_df = self._graph_to_raw_dataset(g)
            self.save_processed(data_df, processed_name)

    @abstractmethod
    def _download_required_data(self) -> str:
        """
        Downloads the required raw data.

        Returns:
            str: Path to the downloaded data.
        """
        pass

    @abstractmethod
    def _extract_class_hierarchy(self, data_path: str) -> "nx.DiGraph":
        """
        Extracts the class hierarchy from the data.
        Constructs a directed graph (DiGraph) using NetworkX, where nodes are annotated with fields/terms from
        the term documents.

        Args:
            data_path (str): Path to the data.

        Returns:
            nx.DiGraph: The class hierarchy graph.
        """
        pass

    @abstractmethod
    def _graph_to_raw_dataset(self, graph: "nx.DiGraph") -> pd.DataFrame:
        """
        Converts the graph to a raw dataset.
        Uses the graph created by `_extract_class_hierarchy` method to extract the
        raw data in Dataframe format with additional columns corresponding to each multi-label class.

        Args:
            graph (nx.DiGraph): The class hierarchy graph.

        Returns:
            pd.DataFrame: The raw dataset.
        """
        pass

    @abstractmethod
    def select_classes(self, g: "nx.DiGraph", *args, **kwargs) -> List:
        """
        Selects classes from the dataset based on a specified criteria.

        Args:
            g (nx.Graph): The graph representing the dataset.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            List: A sorted list of node IDs that meet the specified criteria.
        """
        pass

    def save_processed(self, data: pd.DataFrame, filename: str) -> None:
        """
        Save the processed dataset to a pickle file.

        Args:
            data (pd.DataFrame): The processed dataset to be saved.
            filename (str): The filename for the pickle file.
        """
        pd.to_pickle(data, open(os.path.join(self.processed_dir_main, filename), "wb"))

    def get_processed_pickled_df_file(self, filename: str) -> Optional[pd.DataFrame]:
        """
        Gets the processed dataset pickle file.

        Args:
            filename (str): The filename for the pickle file.

        Returns:
            pd.DataFrame: The processed dataset as a DataFrame.
        """
        file_path = Path(self.processed_dir_main) / filename
        if file_path.exists():
            return pd.read_pickle(file_path)
        return None

    # ------------------------------ Phase: Setup data -----------------------------------
    def setup_processed(self) -> None:
        """
        Transforms `data.pkl` into a model input data format (`data.pt`), ensuring that the data is in a format
        compatible for input to the model.
        The transformed data contains the following keys: `ident`, `features`, `labels`, and `group`.
        This method uses a subclass of Data Reader to perform the transformation.

        Returns:
            None
        """
        os.makedirs(self.processed_dir, exist_ok=True)
        transformed_file_name = self.processed_file_names_dict["data"]
        print(
            f"Missing transformed data (`{transformed_file_name}` file). Transforming data.... "
        )
        torch.save(
            self._load_data_from_file(
                os.path.join(
                    self.processed_dir_main,
                    self.processed_main_file_names_dict["data"],
                )
            ),
            os.path.join(self.processed_dir, transformed_file_name),
        )

    @staticmethod
    def _get_data_size(input_file_path: str) -> int:
        """
        Get the size of the data from a pickled file.

        Args:
            input_file_path (str): The path to the file.

        Returns:
            int: The size of the data.
        """
        with open(input_file_path, "rb") as f:
            df = pd.read_pickle(f)
            print(f"Processed data size ({input_file_path}): {len(df)} rows")
            return len(df)

    @abstractmethod
    def _load_dict(self, input_file_path: str) -> Generator[Dict[str, Any], None, None]:
        """
        Loads data from given pickled file and yields individual dictionaries for each row.

        This method is used by `_load_data_from_file` to generate dictionaries that are then
        processed and converted into a list of dictionaries containing the features and labels.

        Args:
            input_file_path (str): The path to the pickled input file.

        Yields:
            Generator[Dict[str, Any], None, None]: Generator yielding dictionaries.

        """
        pass

    # ------------------------------ Phase: Dynamic Splits -----------------------------------
    @property
    def dynamic_split_dfs(self) -> Dict[str, pd.DataFrame]:
        """
        Property to retrieve dynamic train, validation, and test splits.

        This property checks if dynamic data splits (`_dynamic_df_train`, `_dynamic_df_val`, `_dynamic_df_test`)
        are already loaded. If any of them is None, it either generates them dynamically or retrieves them
        from data file with help of pre-existing split csv file (`splits_file_path`) containing splits assignments.

        Returns:
            dict: A dictionary containing the dynamic train, validation, and test DataFrames.
                Keys are 'train', 'validation', and 'test'.
        """
        if any(
            split is None
            for split in [
                self._dynamic_df_test,
                self._dynamic_df_val,
                self._dynamic_df_train,
            ]
        ):
            if self.splits_file_path is None:
                # Generate splits based on given seed, create csv file to records the splits
                self._generate_dynamic_splits()
            else:
                # If user has provided splits file path, use it to get the splits from the data
                self._retrieve_splits_from_csv()
        return {
            "train": self._dynamic_df_train,
            "validation": self._dynamic_df_val,
            "test": self._dynamic_df_test,
        }

    def _generate_dynamic_splits(self) -> None:
        """
        Generate data splits during runtime and save them in class variables.

        This method loads encoded data and generates train, validation, and test splits based on the loaded data.
        """
        print("\nGenerate dynamic splits...")
        df_train, df_val, df_test = self._get_data_splits()

        # Generate splits.csv file to store ids of each corresponding split
        split_assignment_list: List[pd.DataFrame] = [
            pd.DataFrame({"id": df_train["ident"], "split": "train"}),
            pd.DataFrame({"id": df_val["ident"], "split": "validation"}),
            pd.DataFrame({"id": df_test["ident"], "split": "test"}),
        ]

        combined_split_assignment = pd.concat(split_assignment_list, ignore_index=True)
        combined_split_assignment.to_csv(
            os.path.join(self.processed_dir_main, "splits.csv"), index=False
        )

        # Store the splits in class variables
        self._dynamic_df_train = df_train
        self._dynamic_df_val = df_val
        self._dynamic_df_test = df_test

    @abstractmethod
    def _get_data_splits(self) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        """
        Retrieve the train, validation, and test data splits for the dataset.

        This method returns data splits according to specific criteria implemented
        in the subclasses.

        Returns:
            tuple: A tuple containing DataFrames for train, validation, and test splits.
        """
        pass

    def get_test_split(
        self, df: pd.DataFrame, seed: Optional[int] = None
    ) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Split the input DataFrame into training and testing sets based on multilabel stratified sampling.

        This method uses MultilabelStratifiedShuffleSplit to split the data such that the distribution of labels
        in the training and testing sets is approximately the same. The split is based on the "labels" column
        in the DataFrame.

        Args:
            df (pd.DataFrame): The input DataFrame containing the data to be split. It must contain a column
                               named "labels" with the multilabel data.
            seed (int, optional): The random seed to be used for reproducibility. Default is None.

        Returns:
            Tuple[pd.DataFrame, pd.DataFrame]: A tuple containing the training set and testing set DataFrames.

        Raises:
            ValueError: If the DataFrame does not contain a column named "labels".
        """
        from iterstrat.ml_stratifiers import MultilabelStratifiedShuffleSplit
        from sklearn.model_selection import StratifiedShuffleSplit

        print("Get test data split")

        labels_list = df["labels"].tolist()

        if len(labels_list[0]) > 1:
            splitter = MultilabelStratifiedShuffleSplit(
                n_splits=1, test_size=self.test_split, random_state=seed
            )
        else:
            splitter = StratifiedShuffleSplit(
                n_splits=1, test_size=self.test_split, random_state=seed
            )

        train_indices, test_indices = next(splitter.split(labels_list, labels_list))

        df_train = df.iloc[train_indices]
        df_test = df.iloc[test_indices]
        return df_train, df_test

    def get_train_val_splits_given_test(
        self, df: pd.DataFrame, test_df: pd.DataFrame, seed: int = None
    ) -> Union[Dict[str, pd.DataFrame], Tuple[pd.DataFrame, pd.DataFrame]]:
        """
        Split the dataset into train and validation sets, given a test set.
        Use test set (e.g., loaded from another source or generated in get_test_split), to avoid overlap

        Args:
            df (pd.DataFrame): The original dataset.
            test_df (pd.DataFrame): The test dataset.
            seed (int, optional): The random seed to be used for reproducibility. Default is None.

        Returns:
            Union[Dict[str, pd.DataFrame], Tuple[pd.DataFrame, pd.DataFrame]]: A dictionary containing train and
                validation sets if self.use_inner_cross_validation is True, otherwise a tuple containing the train
                and validation DataFrames. The keys are the names of the train and validation sets, and the values
                are the corresponding DataFrames.
        """
        from iterstrat.ml_stratifiers import (
            MultilabelStratifiedKFold,
            MultilabelStratifiedShuffleSplit,
        )
        from sklearn.model_selection import StratifiedShuffleSplit

        print("Split dataset into train / val with given test set")

        test_ids = test_df["ident"].tolist()
        df_trainval = df[~df["ident"].isin(test_ids)]
        labels_list_trainval = df_trainval["labels"].tolist()

        if self.use_inner_cross_validation:
            folds = {}
            kfold = MultilabelStratifiedKFold(
                n_splits=self.inner_k_folds, random_state=seed
            )
            for fold, (train_ids, val_ids) in enumerate(
                kfold.split(
                    labels_list_trainval,
                    labels_list_trainval,
                )
            ):
                df_validation = df_trainval.iloc[val_ids]
                df_train = df_trainval.iloc[train_ids]
                folds[self.raw_file_names_dict[f"fold_{fold}_train"]] = df_train
                folds[self.raw_file_names_dict[f"fold_{fold}_validation"]] = (
                    df_validation
                )

            return folds

        if len(labels_list_trainval[0]) > 1:
            splitter = MultilabelStratifiedShuffleSplit(
                n_splits=1,
                test_size=self.validation_split / (1 - self.test_split),
                random_state=seed,
            )
        else:
            splitter = StratifiedShuffleSplit(
                n_splits=1,
                test_size=self.validation_split / (1 - self.test_split),
                random_state=seed,
            )

        train_indices, validation_indices = next(
            splitter.split(labels_list_trainval, labels_list_trainval)
        )

        df_validation = df_trainval.iloc[validation_indices]
        df_train = df_trainval.iloc[train_indices]
        return df_train, df_validation

    def _retrieve_splits_from_csv(self) -> None:
        """
        Retrieve previously saved data splits from splits.csv file or from provided file path.

        This method loads the splits.csv file located at `self.splits_file_path`.
        It then loads the encoded data (`data.pt`) and filters it based on the IDs retrieved from
        splits.csv to reconstruct the train, validation, and test splits.
        """
        print(f"\nLoading splits from {self.splits_file_path}...")
        splits_df = pd.read_csv(self.splits_file_path)

        filename = self.processed_file_names_dict["data"]
        data = self.load_processed_data_from_file(
            os.path.join(self.processed_dir, filename)
        )
        df_data = pd.DataFrame(data)

        if self.apply_id_filter:
            print(f"Applying ID filter from {self.apply_id_filter}...")
            with open(self.apply_id_filter, "r") as f:
                id_filter = [
                    line["ident"]
                    for line in torch.load(self.apply_id_filter, weights_only=False)
                ]
            df_data = df_data[df_data["ident"].isin(id_filter)]

        if self.apply_label_filter:
            print(f"Applying label filter from {self.apply_label_filter}...")
            with open(self.apply_label_filter, "r") as f:
                label_filter = [line.strip() for line in f]
            with open(os.path.join(self.processed_dir_main, "classes.txt"), "r") as cf:
                classes = [line.strip() for line in cf]
            # reorder labels
            old_labels = np.stack(df_data["labels"])
            label_mapping = [classes.index(lbl) for lbl in label_filter]
            new_labels = old_labels[:, label_mapping]
            df_data["labels"] = list(new_labels)

        train_ids = splits_df[splits_df["split"] == "train"]["id"]
        validation_ids = splits_df[splits_df["split"] == "validation"]["id"]
        test_ids = splits_df[splits_df["split"] == "test"]["id"]

        self._dynamic_df_train = df_data[df_data["ident"].isin(train_ids)]
        self._dynamic_df_val = df_data[df_data["ident"].isin(validation_ids)]
        self._dynamic_df_test = df_data[df_data["ident"].isin(test_ids)]

    # ------------------------------ Phase: DataLoaders -----------------------------------
    def load_processed_data(
        self, kind: Optional[str] = None, filename: Optional[str] = None
    ) -> List[Dict[str, Any]]:
        """
        Loads processed data from a specified dataset type or file.

        This method retrieves processed data based on the dataset type (`kind`) such as "train",
        "val", or "test", or directly from a provided filename. When `kind` is specified, the method
        leverages the `dynamic_split_dfs` property to dynamically generate or retrieve the corresponding
        data splits if they are not already loaded. If both `kind` and `filename` are provided, `filename`
        takes precedence.

        Args:
            kind (str, optional): The type of dataset to load ("train", "val", or "test").
                If `filename` is provided, this argument is ignored. Defaults to None.
            filename (str, optional): The name of the file to load the dataset from.
                If provided, this takes precedence over `kind`. Defaults to None.

        Returns:
            List[Dict[str, Any]]: A list of dictionaries, where each dictionary contains
            the processed data for an individual data point.

        Raises:
            ValueError: If both `kind` and `filename` are None, as one of them is required to load the dataset.
            KeyError: If the specified `kind` does not exist in the `dynamic_split_dfs` property or
                `processed_file_names_dict`, when expected.
            FileNotFoundError: If the file corresponding to the provided `filename` does not exist.
        """
        if kind is None and filename is None:
            raise ValueError(
                "Either kind or filename is required to load the correct dataset, both are None"
            )

        # If both kind and filename are given, use filename
        if kind is not None and filename is None:
            if self.use_inner_cross_validation and kind != "test":
                filename = self.processed_file_names_dict[
                    f"fold_{self.fold_index}_{kind}"
                ]
            else:
                data_df = self.dynamic_split_dfs[kind]
                return data_df.to_dict(orient="records")

        # If filename is provided
        return self.load_processed_data_from_file(filename)

    def load_processed_data_from_file(self, filename):
        return torch.load(os.path.join(filename), weights_only=False)

    # ------------------------------ Phase: Raw Properties -----------------------------------
    @property
    @abstractmethod
    def base_dir(self) -> str:
        """
        Returns the base directory path for storing data.

        Returns:
            str: The path to the base directory.
        """
        pass

    @property
    def processed_dir_main(self) -> str:
        """
        Returns the main directory path where processed data is stored.

        Returns:
            str: The path to the main processed data directory, based on the base directory and the instance's name.
        """
        return os.path.join(
            self.base_dir,
            self._name,
            "processed",
        )

    @property
    def processed_main_file_names_dict(self) -> dict:
        """
        Returns a dictionary mapping processed data file names.

        Returns:
            dict: A dictionary mapping dataset key to their respective file names.
                  For example, {"data": "data.pkl"}.
        """
        return {"data": self._data_pkl_filename}

    @property
    def raw_file_names(self) -> List[str]:
        """
        Returns a list of raw file names.

        Returns:
            List[str]: A list of file names corresponding to the raw data.
        """
        return list(self.raw_file_names_dict.values())

    @property
    def processed_file_names_dict(self) -> dict:
        """
        Returns a dictionary for the processed and tokenized data files.

        Returns:
            dict: A dictionary mapping dataset keys to their respective file names.
                  For example, {"data": "data.pt"}.
        """
        if self.n_token_limit is not None:
            return {"data": f"data_maxlen{self.n_token_limit}.pt"}
        return {"data": "data.pt"}



import os
import pickle as pkl
from typing import Any, Dict, List, Optional

import numpy as np
import torch
import tqdm
from sklearn.exceptions import NotFittedError
from sklearn.linear_model import LogisticRegression as SklearnLogisticRegression

from chebai.models.base import ChebaiBaseNet

LR_MODEL_PATH = os.path.join("models", "LR")


class LogisticRegression(ChebaiBaseNet):
    """
    Logistic Regression model using scikit-learn, wrapped to fit the ChebaiBaseNet interface.
    """

    def __init__(
        self,
        out_dim: int,
        input_dim: int,
        only_predict_classes: Optional[List] = None,
        n_classes=1528,
        **kwargs,
    ):
        super().__init__(out_dim=out_dim, input_dim=input_dim, **kwargs)
        self.models = [
            SklearnLogisticRegression(solver="liblinear") for _ in range(n_classes)
        ]
        # indices of classes (in the dataset used for training) where a model should be trained
        self.only_predict_classes = only_predict_classes

    def forward(self, x: Dict[str, Any], **kwargs) -> torch.Tensor:
        print(
            f"forward called with x[features].shape {x['features'].shape}, self.training {self.training}"
        )
        if self.training:
            self.fit_sklearn(x["features"], x["labels"])
        preds = []
        for model in self.models:
            try:
                p = torch.from_numpy(model.predict(x["features"])).float()
                p = p.to(x["features"].device)
                preds.append(p)
            except NotFittedError:
                preds.append(
                    torch.zeros((x["features"].shape[0]), device=(x["features"].device))
                )
            except AttributeError:
                preds.append(
                    torch.zeros((x["features"].shape[0]), device=(x["features"].device))
                )
        preds = torch.stack(preds, dim=1)
        print(f"preds shape {preds.shape}")
        return preds.squeeze(-1)

    def fit_sklearn(self, X, y):
        """
        Fit the underlying sklearn model. X and y should be numpy arrays.
        """
        for i, model in tqdm.tqdm(enumerate(self.models), desc="Fitting models"):
            import os

            if os.path.exists(os.path.join(LR_MODEL_PATH, f"LR_model_{i}.pkl")):
                print(f"Loading model {i} from file")
                self.models[i] = pkl.load(
                    open(os.path.join(LR_MODEL_PATH, f"LR_model_{i}.pkl"), "rb")
                )
            else:
                if (
                    self.only_predict_classes and i not in self.only_predict_classes
                ):  # only try these classes
                    continue
                try:
                    model.fit(X, y[:, i])
                except ValueError:
                    self.models[i] = PlaceholderModel()
                # dump
                pkl.dump(
                    model, open(os.path.join(LR_MODEL_PATH, f"LR_model_{i}.pkl"), "wb")
                )

    def configure_optimizers(self, **kwargs):
        pass


class PlaceholderModel:
    """Acts like a trained model, but isn't. Use this if training fails and you need a placeholder."""

    def __init__(self, default_prediction=1):
        self.default_prediction = default_prediction

    def predict(self, preds):
        return np.ones(preds.shape[0]) * self.default_prediction



import os
import pickle
from typing import Any, Literal, Sequence

import torch
from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks import BasePredictionWriter


class PredictionWriter(BasePredictionWriter):
    """
    Custom callback for writing predictions to a file at the end of each epoch.

    Args:
        output_dir (str): The directory where prediction files will be saved.
        write_interval (str): When to write predictions. Options are "batch" or "epoch".
    """

    def __init__(
        self,
        output_dir: str,
        write_interval: Literal["batch", "epoch", "batch_and_epoch"],
    ):
        super().__init__(write_interval)
        self.output_dir = output_dir
        self.prediction_file_name = "predictions.pkl"

    def write_on_epoch_end(
        self,
        trainer: Trainer,
        pl_module: LightningModule,
        predictions: Sequence[Any],
        batch_indices: Sequence[Any],
    ) -> None:
        """
        Writes the predictions to a file at the end of the epoch.

        Args:
            trainer (Trainer): The Trainer instance.
            pl_module (LightningModule): The LightningModule instance.
            predictions (Sequence[Any]): Any sequence of predictions for the epoch.
            batch_indices (Sequence[Any]): Any sequence of batch indices.
        """
        results = [
            dict(
                ident=row["data"]["idents"][0],
                predictions=torch.sigmoid(row["output"]["logits"]).numpy(),
                labels=row["labels"][0].numpy() if row["labels"] is not None else None,
            )
            for row in predictions
        ]
        with open(
            os.path.join(self.output_dir, self.prediction_file_name), "wb"
        ) as fout:
            pickle.dump(results, fout)



import logging
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable, Optional, Union

import torch
from lightning.pytorch.core.module import LightningModule
from lightning.pytorch.utilities.rank_zero import rank_zero_info

from chebai.preprocessing.structures import XYData

logging.getLogger("pysmiles").setLevel(logging.CRITICAL)

_MODEL_REGISTRY = dict()


class ChebaiBaseNet(LightningModule, ABC):
    """
    Base class for Chebai neural network models inheriting from PyTorch Lightning's LightningModule.

    Args:
        criterion (torch.nn.Module, optional): The loss criterion for the model. Defaults to None.
        out_dim (int, optional): The output dimension of the model. Defaults to None.
        train_metrics (torch.nn.Module, optional): The metrics to be used during training. Defaults to None.
        val_metrics (torch.nn.Module, optional): The metrics to be used during validation. Defaults to None.
        test_metrics (torch.nn.Module, optional): The metrics to be used during testing. Defaults to None.
        pass_loss_kwargs (bool, optional): Whether to pass loss kwargs to the criterion. Defaults to True.
        optimizer_kwargs (Dict[str, Any], optional): Additional keyword arguments for the optimizer. Defaults to None.
        **kwargs: Additional keyword arguments.

    """

    def __init__(
        self,
        criterion: torch.nn.Module = None,
        out_dim: Optional[int] = None,
        input_dim: Optional[int] = None,
        train_metrics: Optional[torch.nn.Module] = None,
        val_metrics: Optional[torch.nn.Module] = None,
        test_metrics: Optional[torch.nn.Module] = None,
        pass_loss_kwargs: bool = True,
        optimizer_kwargs: Optional[Dict[str, Any]] = None,
        exclude_hyperparameter_logging: Optional[Iterable[str]] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        # super().__init__()
        if exclude_hyperparameter_logging is None:
            exclude_hyperparameter_logging = tuple()
        self.criterion = criterion
        assert out_dim is not None, "out_dim must be specified"
        assert input_dim is not None, "input_dim must be specified"
        self.out_dim = out_dim
        self.input_dim = input_dim
        print(
            f"Input dimension for the model: {self.input_dim}",
            f"Output dimension for the model: {self.out_dim}",
        )

        self.save_hyperparameters(
            ignore=[
                "criterion",
                "train_metrics",
                "val_metrics",
                "test_metrics",
                *exclude_hyperparameter_logging,
            ]
        )

        self.hparams["out_dim"] = out_dim
        self.hparams["input_dim"] = input_dim

        if optimizer_kwargs:
            self.optimizer_kwargs = optimizer_kwargs
        else:
            self.optimizer_kwargs = dict()
        self.train_metrics = train_metrics
        self.validation_metrics = val_metrics
        self.test_metrics = test_metrics
        self.pass_loss_kwargs = pass_loss_kwargs

    def on_load_checkpoint(self, checkpoint: Dict[str, Any]) -> None:
        # avoid errors due to unexpected keys (e.g., if loading checkpoint from a bce model and using it with a
        # different loss)
        if "criterion.base_loss.pos_weight" in checkpoint["state_dict"]:
            del checkpoint["state_dict"]["criterion.base_loss.pos_weight"]
        if "criterion.pos_weight" in checkpoint["state_dict"]:
            del checkpoint["state_dict"]["criterion.pos_weight"]

    def __init_subclass__(cls, **kwargs):
        """
        Automatically registers subclasses in the model registry to prevent duplicates.

        Args:
            **kwargs: Additional keyword arguments.
        """
        if cls.__name__ in _MODEL_REGISTRY:
            raise ValueError(f"Model {cls.__name__} does already exist")
        else:
            _MODEL_REGISTRY[cls.__name__] = cls

    def _get_prediction_and_labels(
        self, data: Dict[str, Any], labels: torch.Tensor, output: torch.Tensor
    ) -> (torch.Tensor, torch.Tensor):
        """
        Gets the predictions and labels from the model output.

        Args:
            data (Dict[str, Any]): The processed batch data.
            labels (torch.Tensor): The true labels.
            output (torch.Tensor): The model output.

        Returns:
            Tuple[torch.Tensor, torch.Tensor]: Predictions and labels.
        """
        # cast labels to int
        return output, labels.to(torch.int) if labels is not None else labels

    def _process_labels_in_batch(self, batch: XYData) -> torch.Tensor:
        """
        Processes the labels in the batch.

        Args:
            batch (XYData): The input batch of data.

        Returns:
            torch.Tensor: The processed labels.
        """
        return batch.y.float()

    def _process_batch(self, batch: XYData, batch_idx: int) -> Dict[str, Any]:
        """
        Processes the batch data.

        Args:
            batch (XYData): The input batch of data.
            batch_idx (int): The index of the current batch.

        Returns:
            Dict[str, Any]: Processed batch data.
        """
        return dict(
            features=batch.x,
            labels=self._process_labels_in_batch(batch),
            model_kwargs=batch.additional_fields["model_kwargs"],
            loss_kwargs=batch.additional_fields["loss_kwargs"],
            idents=batch.additional_fields["idents"],
        )

    def _process_for_loss(
        self,
        model_output: torch.Tensor,
        labels: torch.Tensor,
        loss_kwargs: Dict[str, Any],
    ) -> (torch.Tensor, torch.Tensor, Dict[str, Any]):
        """
        Processes the data for loss computation.

        Args:
            model_output (torch.Tensor): The model output.
            labels (torch.Tensor): The true labels.
            loss_kwargs (Dict[str, Any]): Additional keyword arguments for the loss function.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, Dict[str, Any]]: Model output, labels, and loss kwargs.
        """
        return model_output, labels, loss_kwargs

    def on_train_epoch_start(self) -> None:
        # pass current epoch to datamodule if it has the attribute curr_epoch (for PubChemBatched dataset)
        rank_zero_info(f"Starting epoch {self.current_epoch}")
        if hasattr(self.trainer.datamodule, "curr_epoch"):
            rank_zero_info(f"Setting datamodule.curr_epoch to {self.current_epoch}")
            self.trainer.datamodule.curr_epoch = self.current_epoch

    def training_step(
        self, batch: XYData, batch_idx: int
    ) -> Dict[str, Union[torch.Tensor, Any]]:
        """
        Defines the training step.

        Args:
            batch (XYData): The input batch of data.
            batch_idx (int): The index of the current batch.

        Returns:
            Dict[str, Union[torch.Tensor, Any]]: The result of the training step.
        """
        return self._execute(
            batch, batch_idx, self.train_metrics, prefix="train_", sync_dist=True
        )

    def validation_step(
        self, batch: XYData, batch_idx: int
    ) -> Dict[str, Union[torch.Tensor, Any]]:
        """
        Defines the validation step.

        Args:
            batch (XYData): The input batch of data.
            batch_idx (int): The index of the current batch.

        Returns:
            Dict[str, Union[torch.Tensor, Any]]: The result of the validation step.
        """
        return self._execute(
            batch, batch_idx, self.validation_metrics, prefix="val_", sync_dist=True
        )

    def test_step(
        self, batch: XYData, batch_idx: int
    ) -> Dict[str, Union[torch.Tensor, Any]]:
        """
        Defines the test step.

        Args:
            batch (XYData): The input batch of data.
            batch_idx (int): The index of the current batch.

        Returns:
            Dict[str, Union[torch.Tensor, Any]]: The result of the test step.
        """
        return self._execute(
            batch, batch_idx, self.test_metrics, prefix="test_", sync_dist=True
        )

    def predict_step(
        self, batch: XYData, batch_idx: int, **kwargs
    ) -> Dict[str, Union[torch.Tensor, Any]]:
        """
        Defines the prediction step.

        Args:
            batch (XYData): The input batch of data.
            batch_idx (int): The index of the current batch.
            **kwargs: Additional keyword arguments.

        Returns:
            Dict[str, Union[torch.Tensor, Any]]: The result of the prediction step.
        """
        return self._execute(batch, batch_idx, self.test_metrics, prefix="", log=False)

    def _execute(
        self,
        batch: XYData,
        batch_idx: int,
        metrics: Optional[torch.nn.Module] = None,
        prefix: Optional[str] = "",
        log: Optional[bool] = True,
        sync_dist: Optional[bool] = False,
    ) -> Dict[str, Union[torch.Tensor, Any]]:
        """
        Executes the model on a batch of data and returns the model output and predictions.

        Args:
            batch (XYData): The input batch of data.
            batch_idx (int): The index of the current batch.
            metrics (torch.nn.Module): A dictionary of metrics to track.
            prefix (str, optional): A prefix to add to the metric names. Defaults to "".
            log (bool, optional): Whether to log the metrics. Defaults to True.
            sync_dist (bool, optional): Whether to synchronize distributed training. Defaults to False.

        Returns:
            Dict[str, Union[torch.Tensor, Any]]: A dictionary containing the processed data, labels, model output,
            predictions, and loss (if applicable).
        """
        assert isinstance(batch, XYData)
        batch = batch.to(self.device)
        data = self._process_batch(batch, batch_idx)
        labels = data["labels"]
        model_output = self(data, **data.get("model_kwargs", dict()))
        pr, tar = self._get_prediction_and_labels(data, labels, model_output)
        d = dict(data=data, labels=labels, output=model_output, preds=pr)
        if log:
            if self.criterion is not None:
                loss_data, loss_labels, loss_kwargs_candidates = self._process_for_loss(
                    model_output, labels, data.get("loss_kwargs", dict())
                )
                loss_kwargs = dict()
                if self.pass_loss_kwargs:
                    loss_kwargs = loss_kwargs_candidates
                loss = self.criterion(loss_data, loss_labels, **loss_kwargs)
                if isinstance(loss, tuple):
                    unnamed_loss_index = 1
                    if isinstance(loss[1], dict):
                        unnamed_loss_index = 2
                        for key, value in loss[1].items():
                            self.log(
                                key,
                                value if isinstance(value, int) else value.item(),
                                batch_size=len(batch),
                                on_step=True,
                                on_epoch=True,
                                prog_bar=False,
                                logger=True,
                                sync_dist=sync_dist,
                            )
                    loss_additional = loss[unnamed_loss_index:]
                    for i, loss_add in enumerate(loss_additional):
                        self.log(
                            f"{prefix}loss_{i}",
                            loss_add if isinstance(loss_add, int) else loss_add.item(),
                            batch_size=len(batch),
                            on_step=True,
                            on_epoch=True,
                            prog_bar=False,
                            logger=True,
                            sync_dist=sync_dist,
                        )
                    loss = loss[0]

                d["loss"] = loss
                self.log(
                    f"{prefix}loss",
                    loss.item(),
                    batch_size=len(batch),
                    on_step=True,
                    on_epoch=True,
                    prog_bar=True,
                    logger=True,
                    sync_dist=sync_dist,
                )
            if metrics and labels is not None:
                for metric_name, metric in metrics.items():
                    metric.update(pr, tar)
                self._log_metrics(prefix, metrics, len(batch))
        if isinstance(d, dict) and "loss" not in d:
            print(f"d has keys {d.keys()}, log={log}, criterion={self.criterion}")
        return d

    def _log_metrics(self, prefix: str, metrics: torch.nn.Module, batch_size: int):
        """
        Logs the metrics for the given prefix.

        Args:
            prefix (str): The prefix to be added to the metric names.
            metrics (torch.nn.Module): A dictionary containing the metrics to be logged.
            batch_size (int): The batch size used for logging.

        Returns:
            None
        """
        # don't use sync_dist=True if the metric is a torchmetrics-metric
        # (see https://github.com/Lightning-AI/pytorch-lightning/discussions/6501#discussioncomment-569757)
        for metric_name, metric in metrics.items():
            m = None  # m = metric.compute()
            if isinstance(m, dict):
                # todo: is this case needed? it requires logging values directly which does not give accurate results
                # with the current metric-setup
                for k, m2 in m.items():
                    self.log(
                        f"{prefix}{metric_name}{k}",
                        m2,
                        batch_size=batch_size,
                        on_step=False,
                        on_epoch=True,
                        prog_bar=True,
                        logger=True,
                    )
            else:
                self.log(
                    f"{prefix}{metric_name}",
                    metric,
                    batch_size=batch_size,
                    on_step=False,
                    on_epoch=True,
                    prog_bar=True,
                    logger=True,
                )

    @abstractmethod
    def forward(self, x: Dict[str, Any]) -> torch.Tensor:
        """
        Defines the forward pass.

        Args:
            x (Dict[str, Any]): The input data.

        Returns:
            torch.Tensor: The model output.
        """
        pass

    def configure_optimizers(self, **kwargs) -> torch.optim.Optimizer:
        """
        Configures the optimizers.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            torch.optim.Optimizer: The optimizer.
        """
        return torch.optim.Adamax(self.parameters(), **self.optimizer_kwargs)



import importlib
import os
import shutil
from pathlib import Path
from typing import Optional, Tuple

import torch
import tqdm
import wandb
import wandb.util as wandb_util
import yaml

from chebai.models.base import ChebaiBaseNet
from chebai.preprocessing.datasets.base import XYBaseDataModule
from chebai.preprocessing.datasets.chebi import _ChEBIDataExtractor


def get_checkpoint_from_wandb(
    epoch: int,
    run: wandb.apis.public.Run,
    root: str = os.path.join("logs", "downloaded_ckpts"),
):
    """
    Gets a wandb checkpoint based on run and epoch, downloads it if necessary.

    Args:
        epoch: The epoch number of the checkpoint to retrieve.
        run: The wandb run object.
        root: The root directory to save the downloaded checkpoint.

    Returns:
        The location of the downloaded checkpoint.
    """
    api = wandb.Api()

    files = run.files()
    for file in files:
        if file.name.startswith(
            f"checkpoints/per_epoch={epoch}"
        ) or file.name.startswith(f"checkpoints/best_epoch={epoch}"):
            dest_path = os.path.join(
                root, run.id, file.name.split("/")[-1].split("_")[1] + ".ckpt"
            )
            # legacy: also look for ckpts in the old format
            old_dest_path = os.path.join(root, run.name, file.name.split("/")[-1])
            if not os.path.isfile(dest_path):
                if os.path.isfile(old_dest_path):
                    print(f"Copying checkpoint from {old_dest_path} to {dest_path}")
                    os.makedirs(os.path.dirname(dest_path), exist_ok=True)
                    shutil.copy2(old_dest_path, dest_path)
                else:
                    print(f"Downloading checkpoint to {dest_path}")
                    wandb_util.download_file_from_url(dest_path, file.url, api.api_key)
            return dest_path
    print(f"No model found for epoch {epoch}")
    return None


def _run_batch(batch, model, collate):
    collated = collate(batch)
    collated.x = collated.to_x(model.device)
    if collated.y is not None:
        collated.y = collated.to_y(model.device)
    processable_data = model._process_batch(collated, 0)
    # del processable_data["loss_kwargs"]
    model_output = model(processable_data, **processable_data["model_kwargs"])
    preds, labels = model._get_prediction_and_labels(
        processable_data, processable_data["labels"], model_output
    )
    return preds, labels


def _run_batch_give_attention(batch, model, collate):
    collated = collate(batch)
    collated.x = collated.to_x(model.device)
    if collated.y is not None:
        collated.y = collated.to_y(model.device)
    processable_data = model._process_batch(collated, 0)
    # del processable_data["loss_kwargs"]
    model_output = model(processable_data, **processable_data["model_kwargs"])
    preds, labels = model._get_prediction_and_labels(
        processable_data, processable_data["labels"], model_output
    )
    return preds, labels, model_output


def _concat_tuple(l_):
    if isinstance(l_[0], tuple):
        print(l_[0])
        return tuple([torch.cat([t[i] for t in l_]) for i in range(len(l_[0]))])
    return torch.cat(l_)


def evaluate_model(
    model: ChebaiBaseNet,
    data_module: XYBaseDataModule,
    filename: Optional[str] = None,
    buffer_dir: Optional[str] = None,
    batch_size: int = 32,
    skip_existing_preds: bool = False,
    kind: str = "test",
) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
    """
    Runs a classification model on the test set of the data module or on the dataset found in the specified file.
    If buffer_dir is set, results will be saved in buffer_dir.

    Note:
        No need to provide "filename" parameter for Chebi dataset, "kind" parameter should be provided.

    Args:
        model: The model to evaluate.
        data_module: The data module containing the dataset.
        filename: Optional file name for the dataset.
        buffer_dir: Optional directory to save the results.
        batch_size: The batch size for evaluation.
        skip_existing_preds: Whether to skip evaluation if predictions already exist.
        kind: Kind of split of the data to be used for testing the model. Default is `test`.

    Returns:
        Tensors with predictions and labels.
    """
    assert model.model_type == "classification"
    model.eval()
    collate = data_module.reader.COLLATOR()

    if isinstance(data_module, _ChEBIDataExtractor):
        # As the dynamic split change is implemented only for chebi-dataset as of now
        data_df = data_module.dynamic_split_dfs[kind]
        data_list = data_df.to_dict(orient="records")
    else:
        data_list = data_module.load_processed_data("test", filename)
    data_list = data_list[: data_module.data_limit]
    preds_list = []
    labels_list = []
    if buffer_dir is not None:
        os.makedirs(buffer_dir, exist_ok=True)
    save_ind = 0
    save_batch_size = 128
    n_saved = 1

    print("")
    for i in tqdm.tqdm(range(0, len(data_list), batch_size)):
        if not (
            skip_existing_preds
            and os.path.isfile(os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"))
        ):
            preds, labels = _run_batch(data_list[i : i + batch_size], model, collate)
            preds_list.append(preds)
            labels_list.append(labels)

            if buffer_dir is not None:
                if n_saved * batch_size >= save_batch_size:
                    torch.save(
                        _concat_tuple(preds_list),
                        os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"),
                    )
                    if labels_list[0] is not None:
                        torch.save(
                            _concat_tuple(labels_list),
                            os.path.join(buffer_dir, f"labels{save_ind:03d}.pt"),
                        )
                    preds_list = []
                    labels_list = []
        if n_saved * batch_size >= save_batch_size:
            save_ind += 1
            n_saved = 0
        n_saved += 1

    if buffer_dir is None:
        test_preds = _concat_tuple(preds_list)
        if labels_list is not None:
            test_labels = _concat_tuple(labels_list)
            return test_preds, test_labels
        return test_preds, None
    elif len(preds_list) > 0:
        if preds_list[0] is not None:
            torch.save(
                _concat_tuple(preds_list),
                os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"),
            )
        if labels_list[0] is not None:
            torch.save(
                _concat_tuple(labels_list),
                os.path.join(buffer_dir, f"labels{save_ind:03d}.pt"),
            )
    return torch.cat(preds_list), torch.cat(labels_list)


def evaluate_model_regression(
    model: ChebaiBaseNet,
    data_module: XYBaseDataModule,
    filename: Optional[str] = None,
    buffer_dir: Optional[str] = None,
    batch_size: int = 32,
    skip_existing_preds: bool = False,
    kind: str = "test",
) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
    """
    Runs a regression model on the test set of the data module or on the dataset found in the specified file.
    If buffer_dir is set, results will be saved in buffer_dir.

    Note:
        No need to provide "filename" parameter for Chebi dataset, "kind" parameter should be provided.

    Args:
        model: The model to evaluate.
        data_module: The data module containing the dataset.
        filename: Optional file name for the dataset.
        buffer_dir: Optional directory to save the results.
        batch_size: The batch size for evaluation.
        skip_existing_preds: Whether to skip evaluation if predictions already exist.
        kind: Kind of split of the data to be used for testing the model. Default is `test`.

    Returns:
        Tensors with predictions and labels.
    """
    model.eval()
    collate = data_module.reader.COLLATOR()

    if isinstance(data_module, _ChEBIDataExtractor):
        # As the dynamic split change is implemented only for chebi-dataset as of now
        data_df = data_module.dynamic_split_dfs[kind]
        data_list = data_df.to_dict(orient="records")
    else:
        data_list = data_module.load_processed_data("test", filename)
    data_list = data_list[: data_module.data_limit]
    preds_list = []
    labels_list = []
    preds_list_all = []
    labels_list_all = []
    if buffer_dir is not None:
        os.makedirs(buffer_dir, exist_ok=True)
    save_ind = 0
    save_batch_size = 128
    n_saved = 1

    print("")
    for i in tqdm.tqdm(range(0, len(data_list), batch_size)):
        if not (
            skip_existing_preds
            and os.path.isfile(os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"))
        ):
            preds, labels = _run_batch(data_list[i : i + batch_size], model, collate)
            preds_list.append(preds)
            labels_list.append(labels)
            preds_list_all.append(preds)
            labels_list_all.append(labels)
            if buffer_dir is not None:
                if n_saved * batch_size >= save_batch_size:
                    torch.save(
                        _concat_tuple(preds_list),
                        os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"),
                    )
                    if labels_list[0] is not None:
                        torch.save(
                            _concat_tuple(labels_list),
                            os.path.join(buffer_dir, f"labels{save_ind:03d}.pt"),
                        )
                    preds_list = []
                    labels_list = []
        if n_saved * batch_size >= save_batch_size:
            save_ind += 1
            n_saved = 0
        n_saved += 1

    if buffer_dir is None:
        test_preds = _concat_tuple(preds_list)
        if labels_list is not None:
            test_labels = _concat_tuple(labels_list)

            return test_preds, test_labels
        return test_preds, None
    else:
        torch.save(
            _concat_tuple(preds_list),
            os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"),
        )
        if labels_list[0] is not None:
            torch.save(
                _concat_tuple(labels_list),
                os.path.join(buffer_dir, f"labels{save_ind:03d}.pt"),
            )
    return torch.cat(preds_list_all), torch.cat(labels_list_all)


def evaluate_model_regression_attention(
    model: ChebaiBaseNet,
    data_module: XYBaseDataModule,
    filename: Optional[str] = None,
    buffer_dir: Optional[str] = None,
    batch_size: int = 32,
    skip_existing_preds: bool = False,
    kind: str = "test",
) -> Tuple[torch.Tensor, Optional[torch.Tensor], list, list]:
    """
    Runs the model on the test set of the data module or on the dataset found in the specified file.
    If buffer_dir is set, results will be saved in buffer_dir.

    Note:
        No need to provide "filename" parameter for Chebi dataset, "kind" parameter should be provided.

    Args:
        model: The model to evaluate.
        data_module: The data module containing the dataset.
        filename: Optional file name for the dataset.
        buffer_dir: Optional directory to save the results.
        batch_size: The batch size for evaluation.
        skip_existing_preds: Whether to skip evaluation if predictions already exist.
        kind: Kind of split of the data to be used for testing the model. Default is `test`.

    Returns:
        Tensors with predictions and labels.
    """
    model.eval()
    collate = data_module.reader.COLLATOR()

    if isinstance(data_module, _ChEBIDataExtractor):
        # As the dynamic split change is implemented only for chebi-dataset as of now
        data_df = data_module.dynamic_split_dfs[kind]
        data_list = data_df.to_dict(orient="records")
    else:
        data_list = data_module.load_processed_data("test", filename)
    data_list = data_list[: data_module.data_limit]
    preds_list = []
    labels_list = []
    preds_list_all = []
    labels_list_all = []
    features_list_all = []
    attention_list_all = []
    if buffer_dir is not None:
        os.makedirs(buffer_dir, exist_ok=True)
    save_ind = 0
    save_batch_size = 128
    n_saved = 1

    print("")
    for i in tqdm.tqdm(range(0, len(data_list), batch_size)):
        if not (
            skip_existing_preds
            and os.path.isfile(os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"))
        ):
            preds, labels, model_output = _run_batch_give_attention(
                data_list[i : i + batch_size], model, collate
            )
            preds_list.append(preds)
            labels_list.append(labels)
            preds_list_all.append(preds)
            labels_list_all.append(labels)
            attention_list_all.append(model_output)
            features_list_all.append(data_list[i : i + batch_size])
            if buffer_dir is not None:
                if n_saved * batch_size >= save_batch_size:
                    torch.save(
                        _concat_tuple(preds_list),
                        os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"),
                    )
                    if labels_list[0] is not None:
                        torch.save(
                            _concat_tuple(labels_list),
                            os.path.join(buffer_dir, f"labels{save_ind:03d}.pt"),
                        )
                    preds_list = []
                    labels_list = []
        if n_saved * batch_size >= save_batch_size:
            save_ind += 1
            n_saved = 0
        n_saved += 1

    if buffer_dir is None:
        test_preds = _concat_tuple(preds_list)
        if labels_list is not None:
            test_labels = _concat_tuple(labels_list)

            return test_preds, test_labels, features_list_all, attention_list_all
        return test_preds, None
    else:
        torch.save(
            _concat_tuple(preds_list),
            os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"),
        )
        if labels_list[0] is not None:
            torch.save(
                _concat_tuple(labels_list),
                os.path.join(buffer_dir, f"labels{save_ind:03d}.pt"),
            )
    return torch.cat(preds_list_all), torch.cat(labels_list_all)


# def evaluate_model_regression(
#     model: ChebaiBaseNet,
#     data_module: XYBaseDataModule,
#     filename: Optional[str] = None,
#     buffer_dir: Optional[str] = None,
#     batch_size: int = 32,
#     skip_existing_preds: bool = False,
#     kind: str = "test",
# ) -> Tuple[torch.Tensor, Optional[torch.Tensor]]:
#     """
#     Runs the model on the test set of the data module or on the dataset found in the specified file.
#     If buffer_dir is set, results will be saved in buffer_dir.

#     Note:
#         No need to provide "filename" parameter for Chebi dataset, "kind" parameter should be provided.

#     Args:
#         model: The model to evaluate.
#         data_module: The data module containing the dataset.
#         filename: Optional file name for the dataset.
#         buffer_dir: Optional directory to save the results.
#         batch_size: The batch size for evaluation.
#         skip_existing_preds: Whether to skip evaluation if predictions already exist.
#         kind: Kind of split of the data to be used for testing the model. Default is `test`.

#     Returns:
#         Tensors with predictions and labels.
#     """
#     model.eval()
#     collate = data_module.reader.COLLATOR()

#     if isinstance(data_module, _ChEBIDataExtractor):
#         # As the dynamic split change is implemented only for chebi-dataset as of now
#         data_df = data_module.dynamic_split_dfs[kind]
#         data_list = data_df.to_dict(orient="records")
#     else:
#         data_list = data_module.load_processed_data("test", filename)
#     data_list = data_list[: data_module.data_limit]
#     preds_list = []
#     labels_list = []
#     preds_list_all = []
#     labels_list_all = []
#     if buffer_dir is not None:
#         os.makedirs(buffer_dir, exist_ok=True)
#     save_ind = 0
#     save_batch_size = 128
#     n_saved = 1

#     print("")
#     for i in tqdm.tqdm(range(0, len(data_list), batch_size)):
#         if not (
#             skip_existing_preds
#             and os.path.isfile(os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"))
#         ):
#             preds, labels = _run_batch(data_list[i : i + batch_size], model, collate)
#             preds_list.append(preds)
#             labels_list.append(labels)
#             preds_list_all.append(preds)
#             labels_list_all.append(labels)
#             if buffer_dir is not None:
#                 if n_saved * batch_size >= save_batch_size:
#                     torch.save(
#                         _concat_tuple(preds_list),
#                         os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"),
#                     )
#                     if labels_list[0] is not None:
#                         torch.save(
#                             _concat_tuple(labels_list),
#                             os.path.join(buffer_dir, f"labels{save_ind:03d}.pt"),
#                         )
#                     preds_list = []
#                     labels_list = []
#         if n_saved * batch_size >= save_batch_size:
#             save_ind += 1
#             n_saved = 0
#         n_saved += 1

#     if buffer_dir is None:
#         test_preds = _concat_tuple(preds_list)
#         if labels_list is not None:
#             test_labels = _concat_tuple(labels_list)

#             return test_preds, test_labels
#         return test_preds, None
#     else:
#         torch.save(
#             _concat_tuple(preds_list),
#             os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"),
#         )
#         if labels_list[0] is not None:
#             torch.save(
#                 _concat_tuple(labels_list),
#                 os.path.join(buffer_dir, f"labels{save_ind:03d}.pt"),
#             )
#     return torch.cat(preds_list_all), torch.cat(labels_list_all)


# def evaluate_model_regression_attention(
#     model: ChebaiBaseNet,
#     data_module: XYBaseDataModule,
#     filename: Optional[str] = None,
#     buffer_dir: Optional[str] = None,
#     batch_size: int = 32,
#     skip_existing_preds: bool = False,
#     kind: str = "test",
# ) -> Tuple[torch.Tensor, Optional[torch.Tensor], list, list]:
#     """
#     Runs the model on the test set of the data module or on the dataset found in the specified file.
#     If buffer_dir is set, results will be saved in buffer_dir.

#     Note:
#         No need to provide "filename" parameter for Chebi dataset, "kind" parameter should be provided.

#     Args:
#         model: The model to evaluate.
#         data_module: The data module containing the dataset.
#         filename: Optional file name for the dataset.
#         buffer_dir: Optional directory to save the results.
#         batch_size: The batch size for evaluation.
#         skip_existing_preds: Whether to skip evaluation if predictions already exist.
#         kind: Kind of split of the data to be used for testing the model. Default is `test`.

#     Returns:
#         Tensors with predictions and labels.
#     """
#     model.eval()
#     collate = data_module.reader.COLLATOR()

#     if isinstance(data_module, _ChEBIDataExtractor):
#         # As the dynamic split change is implemented only for chebi-dataset as of now
#         data_df = data_module.dynamic_split_dfs[kind]
#         data_list = data_df.to_dict(orient="records")
#     else:
#         data_list = data_module.load_processed_data("test", filename)
#     data_list = data_list[: data_module.data_limit]
#     preds_list = []
#     labels_list = []
#     preds_list_all = []
#     labels_list_all = []
#     features_list_all = []
#     attention_list_all = []
#     if buffer_dir is not None:
#         os.makedirs(buffer_dir, exist_ok=True)
#     save_ind = 0
#     save_batch_size = 128
#     n_saved = 1

#     print("")
#     for i in tqdm.tqdm(range(0, len(data_list), batch_size)):
#         if not (
#             skip_existing_preds
#             and os.path.isfile(os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"))
#         ):
#             preds, labels, model_output = _run_batch_give_attention(
#                 data_list[i : i + batch_size], model, collate
#             )
#             preds_list.append(preds)
#             labels_list.append(labels)
#             preds_list_all.append(preds)
#             labels_list_all.append(labels)
#             attention_list_all.append(model_output)
#             features_list_all.append(data_list[i : i + batch_size])
#             if buffer_dir is not None:
#                 if n_saved * batch_size >= save_batch_size:
#                     torch.save(
#                         _concat_tuple(preds_list),
#                         os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"),
#                     )
#                     if labels_list[0] is not None:
#                         torch.save(
#                             _concat_tuple(labels_list),
#                             os.path.join(buffer_dir, f"labels{save_ind:03d}.pt"),
#                         )
#                     preds_list = []
#                     labels_list = []
#         if n_saved * batch_size >= save_batch_size:
#             save_ind += 1
#             n_saved = 0
#         n_saved += 1

#     if buffer_dir is None:
#         test_preds = _concat_tuple(preds_list)
#         if labels_list is not None:
#             test_labels = _concat_tuple(labels_list)

#             return test_preds, test_labels, features_list_all, attention_list_all
#         return test_preds, None
#     else:
#         torch.save(
#             _concat_tuple(preds_list),
#             os.path.join(buffer_dir, f"preds{save_ind:03d}.pt"),
#         )
#         if labels_list[0] is not None:
#             torch.save(
#                 _concat_tuple(labels_list),
#                 os.path.join(buffer_dir, f"labels{save_ind:03d}.pt"),
#             )
#     return (
#         torch.cat(preds_list_all),
#         torch.cat(labels_list_all),
#         features_list_all,
#         attention_list_all,
#     )


def load_results_from_buffer(
    buffer_dir: str, device: torch.device
) -> Tuple[Optional[torch.Tensor], Optional[torch.Tensor]]:
    """
    Load results stored in evaluate_model() from the buffer directory.

    Args:
        buffer_dir: The directory containing the buffered results.
        device: The device to load the results onto.

    Returns:
        Tensors with predictions and labels.
    """
    preds_list = []
    labels_list = []

    i = 0
    filename = f"preds{i:03d}.pt"
    while os.path.isfile(os.path.join(buffer_dir, filename)):
        preds_list.append(
            torch.load(
                os.path.join(buffer_dir, filename),
                map_location=torch.device(device),
                weights_only=False,
            )
        )
        i += 1
        filename = f"preds{i:03d}.pt"

    i = 0
    filename = f"labels{i:03d}.pt"
    while os.path.isfile(os.path.join(buffer_dir, filename)):
        labels_list.append(
            torch.load(
                os.path.join(buffer_dir, filename),
                map_location=torch.device(device),
                weights_only=False,
            )
        )
        i += 1
        filename = f"labels{i:03d}.pt"

    if len(preds_list) > 0:
        test_preds = torch.cat(preds_list)
    else:
        test_preds = None
    if len(labels_list) > 0:
        test_labels = torch.cat(labels_list)
    else:
        test_labels = None

    return test_preds, test_labels


def load_class(class_path: str) -> type:
    module_path, class_name = class_path.rsplit(".", 1)
    module = importlib.import_module(module_path)
    return getattr(module, class_name)


def load_data_instance(data_cls_path: str, data_cls_kwargs: dict):
    assert isinstance(data_cls_kwargs, dict), "data_cls_kwargs must be a dict"
    data_cls = load_class(data_cls_path)
    assert isinstance(data_cls, type), f"{data_cls} is not a class."
    assert issubclass(
        data_cls, XYBaseDataModule
    ), f"{data_cls} must inherit from XYBaseDataModule"
    return data_cls(**data_cls_kwargs)


def load_model_for_inference(
    model_ckpt_path: str, model_cls_path: str, model_load_kwargs: dict, **kwargs
) -> ChebaiBaseNet:
    """
    Loads a model checkpoint and its label-related properties.

    Returns:
        Tuple[LightningModule, Dict[str, torch.Tensor]]: The model and its label properties.
    """
    assert isinstance(model_load_kwargs, dict), "model_load_kwargs must be a dict"

    model_name = kwargs.get("model_name", model_ckpt_path)

    if not Path(model_ckpt_path).exists():
        raise FileNotFoundError(
            f"Model path '{model_ckpt_path}' for '{model_name}' does not exist."
        )

    lightning_cls = load_class(model_cls_path)

    assert isinstance(lightning_cls, type), f"{lightning_cls} is not a class."
    assert issubclass(
        lightning_cls, ChebaiBaseNet
    ), f"{lightning_cls} must inherit from ChebaiBaseNet"
    try:
        model = lightning_cls.load_from_checkpoint(model_ckpt_path, **model_load_kwargs)
    except Exception as e:
        raise RuntimeError(f"Error loading model {model_name} \n Error: {e}") from e

    assert isinstance(
        model, ChebaiBaseNet
    ), f"Model: {model}(Model Name: {model_name}) is not a ChebaiBaseNet instance."
    model.eval()
    model.freeze()
    return model


def parse_config_file(config_path: str) -> tuple[str, dict]:
    path = Path(config_path)

    # Check file existence
    if not path.exists():
        raise FileNotFoundError(f"Config file not found: {config_path}")

    # Check file extension
    if path.suffix.lower() not in [".yml", ".yaml"]:
        raise ValueError(
            f"Unsupported config file type: {path.suffix}. Expected .yaml or .yml"
        )

    # Load YAML content
    with open(path, "r") as f:
        config: dict = yaml.safe_load(f)

    class_path: str = config["class_path"]
    init_args: dict = config.get("init_args", {})
    assert isinstance(init_args, dict), "init_args must be a dictionary"
    return class_path, init_args


if __name__ == "__main__":
    import sys

    buffer_dir = os.path.join("results_buffer", sys.argv[1], "ChEBIOver100_train")
    buffer_dir_concat = os.path.join(
        "results_buffer", "concatenated", sys.argv[1], "ChEBIOver100_train"
    )
    os.makedirs(buffer_dir_concat, exist_ok=True)
    preds, labels = load_results_from_buffer(buffer_dir, "cpu")
    torch.save(preds, os.path.join(buffer_dir_concat, "preds000.pt"))
    torch.save(labels, os.path.join(buffer_dir_concat, "labels000.pt"))



import torch
from torch import Tensor
from torchmetrics.regression import MeanSquaredError

# from chebai.callbacks.epoch_metrics import BalancedAccuracy, MacroF1
# from chebai.result.utils import *

# def visualise_f1(logs_path: str) -> None:
#     """
#     Visualize F1 scores from metrics.csv and save the plot as f1_plot.png.

#     Args:
#         logs_path: The path to the directory containing metrics.csv.
#     """
#     df = pd.read_csv(os.path.join(logs_path, "metrics.csv"))
#     df_loss = df.melt(
#         id_vars="epoch",
#         value_vars=[
#             "val_ep_macro-f1",
#             "val_micro-f1",
#             "train_micro-f1",
#             "train_ep_macro-f1",
#         ],
#     )
#     lineplt = sns.lineplot(df_loss, x="epoch", y="value", hue="variable")
#     plt.savefig(os.path.join(logs_path, "f1_plot.png"))
#     plt.show()


def metrics_regression(
    preds: Tensor,
    labels: Tensor,
    device: torch.device,
    markdown_output: bool = False,
) -> None:
    """
    Prints relevant metrics, including micro and macro F1, recall and precision,
    best k classes, and worst classes.

    Args:
        preds: Predicted labels as a tensor.
        labels: True labels as a tensor.
        device: The device to perform computations on.
        classes: Optional list of class names.
        top_k: The number of top classes to display based on F1 score.
        markdown_output: If True, print metrics in markdown format.
    """
    mse = MeanSquaredError()
    mse = mse.to(labels.device)

    rmse = MeanSquaredError(squared=False)
    rmse = rmse.to(labels.device)

    return (mse(preds, labels), rmse(preds, labels))

    # print(f"Micro-F1: {f1_micro(preds, labels):3f}")
    # print(f"Balanced Accuracy: {my_bal_acc(preds, labels):3f}")

    # if markdown_output:
    #     print(
    #         f"| Model | MSE | RMSE | Macro-Precision | Micro-Precision | Macro-Recall | Micro-Recall | Balanced Accuracy"
    #     )
    #     print(f"| --- | --- | --- | --- | --- | --- | --- | --- |")
    #     print(
    #         f"| Elektra | {my_f1_macro(preds, labels):3f} | {f1_micro(preds, labels):3f} | {precision_macro(preds, labels):3f} | "
    #         f"{precision_micro(preds, labels):3f} | {recall_macro(preds, labels):3f} | "
    #         f"{recall_micro(preds, labels):3f} | {my_bal_acc(preds, labels):3f} |"
    #     )



from chebai.cli import cli

if __name__ == "__main__":
    """
    Entry point for the CLI application.

    This script calls the `cli` function from the `chebai.cli` module
    when executed as the main program.
    """
    cli()



import os

from lightning.fabric.utilities.cloud_io import _is_dir
from lightning.fabric.utilities.types import _PATH
from lightning.pytorch import LightningModule, Trainer
from lightning.pytorch.callbacks.model_checkpoint import ModelCheckpoint
from lightning.pytorch.loggers import WandbLogger
from lightning.pytorch.utilities.rank_zero import rank_zero_info
from lightning_utilities.core.rank_zero import rank_zero_warn


class CustomModelCheckpoint(ModelCheckpoint):
    """
    Custom checkpoint class that resolves checkpoint paths to ensure checkpoints are saved in the same directory
    as other logs when using CustomLogger.
    Inherits from PyTorch Lightning's ModelCheckpoint class.
    """

    def setup(self, trainer: Trainer, pl_module: LightningModule, stage: str) -> None:
        """
        Setup the directory path for saving checkpoints. If the directory path is not set, it resolves the checkpoint
        directory using the custom logger's directory.

        Note:
            Same as in parent class, duplicated to be able to call self.__resolve_ckpt_dir

        Args:
            trainer (Trainer): The Trainer instance.
            pl_module (LightningModule): The LightningModule instance.
            stage (str): The stage of training (e.g., 'fit').
        """
        if self.dirpath is not None:
            self.dirpath = None
        dirpath = self.__resolve_ckpt_dir(trainer)
        dirpath = trainer.strategy.broadcast(dirpath)
        self.dirpath = dirpath
        if trainer.is_global_zero and stage == "fit":
            self.__warn_if_dir_not_empty(self.dirpath)

    def __warn_if_dir_not_empty(self, dirpath: _PATH) -> None:
        """
        Warn if the checkpoint directory is not empty.

        Note:
            Same as in parent class, duplicated because method in parent class is not accessible

        Args:
            dirpath (_PATH): The path to the checkpoint directory.
        """
        if (
            self.save_top_k != 0
            and _is_dir(self._fs, dirpath, strict=True)
            and len(self._fs.ls(dirpath)) > 0
        ):
            rank_zero_warn(f"Checkpoint directory {dirpath} exists and is not empty.")

    def __resolve_ckpt_dir(self, trainer: Trainer) -> _PATH:
        """
        Resolve the checkpoint directory path, ensuring compatibility with WandbLogger by saving checkpoints
        in the same directory as Wandb logs.

        Note:
            Overwritten for compatibility with wandb -> saves checkpoints in same dir as wandb logs

        Args:
            trainer (Trainer): The Trainer instance.

        Returns:
            _PATH: The resolved checkpoint directory path.
        """
        rank_zero_info("Resolving checkpoint dir (custom)")
        if self.dirpath is not None:
            # short circuit if dirpath was passed to ModelCheckpoint
            return self.dirpath
        if len(trainer.loggers) > 0:
            if trainer.loggers[0].save_dir is not None:
                save_dir = trainer.loggers[0].save_dir
            else:
                save_dir = trainer.default_root_dir
            name = trainer.loggers[0].name
            version = trainer.loggers[0].version
            version = version if isinstance(version, str) else f"version_{version}"
            logger = trainer.loggers[0]
            if isinstance(logger, WandbLogger) and isinstance(
                logger.experiment.dir, str
            ):
                ckpt_path = os.path.join(logger.experiment.dir, "checkpoints")
            else:
                ckpt_path = os.path.join(save_dir, str(name), version, "checkpoints")
        else:
            # if no loggers, use default_root_dir
            ckpt_path = os.path.join(trainer.default_root_dir, "checkpoints")

        rank_zero_info(f"Now using checkpoint path {ckpt_path}")
        return ckpt_path



import os
from typing import Any

import torch

# Get the absolute path of the current file's directory
MODULE_PATH = os.path.abspath(os.path.dirname(__file__))


class CustomTensor(torch.Tensor):
    """
    A custom tensor class inheriting from `torch.Tensor`.

    This class allows for the creation of tensors using the provided data.

    Attributes:
        data (Any): The data to be converted into a tensor.
    """

    def __new__(cls, data: Any) -> "CustomTensor":
        """
        Creates a new instance of CustomTensor.

        Args:
            data (Any): The data to be converted into a tensor.

        Returns:
            CustomTensor: A tensor containing the provided data.
        """
        return torch.tensor(data)
